#!/usr/bin/env python
#
# Copyright (c) Facebook, Inc. and its affiliates.
import argparse
import ast
import contextlib
import os
import random
import termios
import time
import timeit
import tty

import gym

import nle  # noqa: F401
from nle import nethack

_ACTIONS = tuple(
    [nethack.MiscAction.MORE]
    + list(nethack.CompassDirection)
    + list(nethack.CompassDirectionLonger)
)


@contextlib.contextmanager
def dummy_context():
    yield None


@contextlib.contextmanager
def no_echo():
    tt = termios.tcgetattr(0)
    try:
        tty.setraw(0)
        yield
    finally:
        termios.tcsetattr(0, termios.TCSAFLUSH, tt)


def go_back(num_lines):
    print("\033[%dA" % num_lines)


def get_action(env, is_raw_env):
    if FLAGS.mode == "random":
        if not is_raw_env:
            action = env.action_space.sample()
        else:
            action = random.choice(_ACTIONS)
            print(action)
    elif FLAGS.mode == "human":
        while True:
            with no_echo():
                ch = ord(os.read(0, 1))
            if ch in [nethack.C("c")]:
                print("Received exit code {}. Aborting.".format(ch))
                return None
            try:
                if is_raw_env:
                    action = ch
                else:
                    action = env.actions.index(ch)
                break
            except ValueError:
                print(
                    ("Selected action '%s' is not in action list. Please try again.")
                    % chr(ch)
                )
                if not FLAGS.print_frames_separately:
                    print("\033[2A")  # Go up 2 lines.
                continue
    return action


def play():
    is_raw_env = FLAGS.env == "raw"

    if is_raw_env:
        if FLAGS.savedir is not None:
            os.makedirs(FLAGS.savedir, exist_ok=True)
            ttyrec = os.path.join(FLAGS.savedir, "nle.ttyrec.bz2")
        else:
            ttyrec = "/dev/null"
        env = nethack.Nethack(ttyrec=ttyrec, wizard=FLAGS.wizard)
    else:
        env = gym.make(
            FLAGS.env,
            save_ttyrec_every=2,
            savedir=FLAGS.savedir,
            max_episode_steps=FLAGS.max_steps,
            allow_all_yn_questions=True,
            allow_all_modes=True,
            wizard=FLAGS.wizard,
        )
        if FLAGS.seeds is not None:
            env.seed(FLAGS.seeds)

    obs = env.reset()

    steps = 0
    episodes = 0
    reward = 0.0
    action = None

    mean_sps = 0
    mean_reward = 0.0

    total_start_time = timeit.default_timer()
    start_time = total_start_time

    while True:
        if not FLAGS.no_render:
            if not is_raw_env:
                print("-" * 8 + " " * 71)
                print(f"Previous reward: {str(reward):64s}")
                act_str = repr(env.actions[action]) if action is not None else ""
                print(f"Previous action: {str(act_str):64s}")
                print("-" * 8)
                env.render(FLAGS.render_mode)
                print("-" * 8)
                print(obs["blstats"])
                if not FLAGS.print_frames_separately:
                    go_back(num_lines=33)
            else:
                print("Previous action:", action)
                obs = dict(zip(nle.nethack.OBSERVATION_DESC.keys(), obs))
                print(
                    nle.nethack.tty_render(
                        obs["tty_chars"], obs["tty_colors"], obs["tty_cursor"]
                    )
                )
                if not FLAGS.print_frames_separately:
                    go_back(num_lines=len(obs["tty_chars"]) + 3)

        action = get_action(env, is_raw_env)

        if action is None:
            break

        if is_raw_env:
            obs, done = env.step(action)
        else:
            obs, reward, done, info = env.step(action)
        steps += 1

        if is_raw_env:
            done = done or steps >= FLAGS.max_steps  # NLE does this by default.
        else:
            mean_reward += (reward - mean_reward) / steps

        if not done:
            continue

        time_delta = timeit.default_timer() - start_time

        if not is_raw_env:
            print("Final reward:", reward)
            print("End status:", info["end_status"].name)
            print("Mean reward:", mean_reward)

        sps = steps / time_delta
        print("Episode: %i. Steps: %i. SPS: %f" % (episodes, steps, sps))

        episodes += 1
        mean_sps += (sps - mean_sps) / episodes

        start_time = timeit.default_timer()

        steps = 0
        mean_reward = 0.0

        if episodes == FLAGS.ngames:
            break
        env.reset()
    env.close()
    print(
        "Finished after %i episodes and %f seconds. Mean sps: %f"
        % (episodes, timeit.default_timer() - total_start_time, mean_sps)
    )


def main():
    parser = argparse.ArgumentParser(description="NLE Play tool.")
    parser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        help="Enables debug mode, which will drop stack into "
        "an ipdb shell if an exception is raised.",
    )
    parser.add_argument(
        "-m",
        "--mode",
        type=str,
        default="human",
        choices=["human", "random"],
        help="Control mode. Defaults to 'human'.",
    )
    parser.add_argument(
        "-e",
        "--env",
        type=str,
        default="NetHackScore-v0",
        help="Gym environment spec. Defaults to 'NetHackStaircase-v0'.",
    )
    parser.add_argument(
        "-n",
        "--ngames",
        type=int,
        default=1,
        help="Number of games to be played before exiting. "
        "NetHack will auto-restart if > 1.",
    )
    parser.add_argument(
        "--max-steps",
        type=int,
        default=1_000_000,
        help="Number of maximum steps per episode.",
    )
    parser.add_argument(
        "--seeds",
        default=None,
        help="Seeds to send to NetHack. Can be a dict or int. "
        "Defaults to None (no seeding).",
    )
    parser.add_argument(
        "--savedir",
        default="nle_data/play_data",
        help="Directory path where data will be saved. "
        "Defaults to 'nle_data/play_data'.",
    )
    parser.add_argument(
        "--no-render", action="store_true", help="Disables env.render()."
    )
    parser.add_argument(
        "--render_mode",
        type=str,
        default="human",
        choices=["human", "full", "ansi"],
        help="Render mode. Defaults to 'human'.",
    )
    parser.add_argument(
        "--print-frames-separately",
        "-p",
        action="store_true",
        help="Don't overwrite frames, print them all.",
    )
    parser.add_argument(
        "--wizard",
        "-D",
        action="store_true",
        help="Use wizard mode.",
    )
    global FLAGS
    FLAGS = parser.parse_args()

    if FLAGS.debug:
        import ipdb

        cm = ipdb.launch_ipdb_on_exception
    else:
        cm = dummy_context

    with cm():
        if FLAGS.seeds is not None:
            # to handle both int and dicts
            FLAGS.seeds = ast.literal_eval(FLAGS.seeds)

        if FLAGS.savedir == "args":
            FLAGS.savedir = "{}_{}_{}.zip".format(
                time.strftime("%Y%m%d-%H%M%S"), FLAGS.mode, FLAGS.env
            )
        elif FLAGS.savedir == "None":
            FLAGS.savedir = None  # Not saving any ttyrecs.

        play()


if __name__ == "__main__":
    main()



from sample_factory.algorithms.utils.arguments import arg_parser
from sample_factory.algorithms.utils.arguments import parse_args


def custom_parse_args(argv=None, evaluation=False):
    parser = arg_parser(argv, evaluation=evaluation)

    # add custom args here
    parser.add_argument(
        "--nle_env_name", type=str, default="NetHackChallenge-v0", help=""
    )
    parser.add_argument(
        "--transformer_hidden_size",
        type=int,
        default=64,
        help="size of transformer hidden layers",
    )
    parser.add_argument(
        "--transformer_hidden_layers",
        type=int,
        default=2,
        help="number of transformer hidden layers",
    )
    parser.add_argument(
        "--transformer_attention_heads",
        type=int,
        default=2,
        help="number of transformer attention heads",
    )
    parser.add_argument(
        "--max_token_length",
        type=int,
        default=256,
        help="Maximum token input length before truncation",
    )

    cfg = parse_args(argv=argv, evaluation=evaluation, parser=parser)
    return cfg



import os

import pytest
import test_converter

import nle.env.tasks
from nle.dataset import db
from nle.dataset import populate_db

TTYRECS = [
    ("aaa", test_converter.TTYREC_2020),
    ("bbb", test_converter.TTYREC_IBMGRAPHICS),
    ("ccc", test_converter.TTYREC_DECGRAPHICS),
]

HOUR = 3600
YEAR = 365 * 24 * HOUR


def gen_ttyrecs(savedir):
    for _ in range(2):
        env = nle.env.tasks.NetHackChallenge(
            savedir=savedir, character="mon-hum-neu-mal", save_ttyrec_every=2
        )
        env.reset()
        for _ in range(5):
            # Need to end naturally to be recorded.
            for c in [ord(" "), ord(" "), ord("<"), ord("y")]:
                _, _, done, *_ = env.step(env.actions.index(c))
            assert done
            env.reset()
        env.close()


@pytest.fixture(scope="session")
def mockdata(tmpdir_factory):  # Create mock data.
    """This fixture needs to be imported to generate a fake db.

    It generates 1 database with 3 datasets in it:

    1. "basictest" - test basic construction of db
        * 3 users "aaa", "bbb", "ccc",
        * each user with 3 copies of the same ttyrec ("[aaa|bbb|ccc].ttyrec.bz2")
        * user "aaa" and "ccc" have the same ttyrecs, "bbb"'s is different
        * user "ccc" has all of their ttyrecs registered as just one "game"/episode
    2. "altorgtest" - test basic construction of db from
            nle/tests/altorg subdirectory
    3. "nletest" - test basic construction of db from data generated by nle
    """
    basetemp = tmpdir_factory.getbasetemp()

    paths = []
    for user, ttyrec in TTYRECS:
        if basetemp.join(user).exists():
            break
        with open(test_converter.getfilename(ttyrec), "rb") as f:
            rec = f.read()
        d = tmpdir_factory.mktemp(user, numbered=False)
        for fn, _ in TTYRECS:
            fn = d.join(fn + ".ttyrec.bz2")
            paths.append(str(fn))
            fn.ensure()
            fn.write(rec, "wb")

    oldcwd = os.getcwd()
    newcwd = tmpdir_factory.getbasetemp()

    try:
        os.chdir(newcwd)
        if not newcwd.join("ttyrecs.db").exists():
            db.create("ttyrecs.db")

            # 1.  Add "basictest" dataset
            with db.db(rw=True) as conn:
                dataset_name = "basictest"
                c = conn.cursor()
                root = str(newcwd)
                paths = [os.path.relpath(p, root) for p in paths]
                games = [1, 2, 3, 4, 5, 6, 7, 7, 7]
                parts = [0, 0, 0, 0, 0, 0, 0, 1, 2]

                c.execute(
                    (
                        "INSERT INTO roots (root, dataset_name, ttyrec_version) "
                        "VALUES (?,?,1)"
                    ),
                    (root, dataset_name),
                )
                c.executemany(
                    "INSERT INTO datasets (gameid, dataset_name) VALUES (?, ?)",
                    zip(range(1, 8), [dataset_name] * 8),
                )
                c.executemany(
                    "INSERT INTO ttyrecs (path, part, gameid, size) VALUES (?,?, ?,?)",
                    zip(paths, parts, games, range(100, 1000, 100)),
                )
                c.executemany(
                    "INSERT INTO games (gameid, name, death, points) VALUES (?,?,?,?)",
                    [
                        (1, "aaa", "met a sticky end", 0),
                        (2, "aaa", "got into a pickle", 1),
                        (3, "aaa", "ran into trouble", 2),
                        (4, "bbb", "met a sticky end", 3),
                        (5, "bbb", "got into a pickle", 4),
                        (6, "bbb", "ran into trouble", 5),
                        (7, "ccc", "ascended", 999),
                    ],
                )
                conn.commit()

            # 2. Add "altorgtest" datase
            populate_db.add_altorg_directory(
                test_converter.getfilename("altorg"), "altorgtest"
            )
            db.set_root("altorgtest", "/path/to/altorg/")

            # 3. Generate and Add "nletest" dataset
            # NB: Uncomment when adding nledata tests
            gen_ttyrecs("")
            populate_db.add_nledata_directory(
                str(os.path.join(root, "nle_data")), "nletest"
            )

        yield tmpdir_factory.getbasetemp()
    finally:
        os.chdir(oldcwd)


@pytest.fixture(scope="session")
def conn(mockdata):
    """This fixture needs to be imported to receive a connection to working db."""
    with db.db() as conn:
        yield conn


class TestDB:
    def test_conn(self, conn):
        assert conn

    def test_ls(self, conn):
        for i, row in enumerate(db.ls(conn)):
            rowid, path = row
            assert int(rowid) == i + 1
        assert i + 1 == 28

    def test_vacuum(self):
        db.vacuum()

    def test_getrow(self):
        row1 = db.get_row("1")
        row5 = db.get_row("5")
        row8 = db.get_row("9")

        assert len(row8) == 6
        rowids, paths, parts, sizes, mtimes, *_ = zip(row1, row5, row8)
        assert rowids == (1, 5, 9)
        # Order of glob (via os.listdir) OS dependent.
        names, ttyrecs = list(zip(*reversed(TTYRECS)))
        suffix = ".ttyrec.bz2"
        for p in paths:
            assert os.path.dirname(p) in names
            assert os.path.basename(p) in [t + suffix for t in names]
        assert sizes == (100, 500, 900)

    def test_getmeta(self, conn):
        with conn:
            meta = db.get_meta(conn)
        ctime, mtime = meta

        assert 0 < mtime - ctime < 45

    def test_setroot(self):
        dataset_name = "basictest"
        oldroot = db.get_root(dataset_name)
        newroot = "/my/new/root"
        db.set_root(dataset_name, newroot)
        assert db.get_root(dataset_name) == newroot
        db.set_root(dataset_name, oldroot)
        assert db.get_root(dataset_name) == oldroot

    def test_games(self, conn):
        user_games = {"aaa": [], "bbb": [], "ccc": []}
        for user, gameid, start, end in db.get_games("basictest"):
            user_games[user].append((gameid, start, end))

        assert len(user_games["aaa"]) == 3
        assert len(user_games["bbb"]) == 3
        assert len(user_games["ccc"]) == 1

        cmd = "SELECT * FROM ttyrecs WHERE gameid = ? ORDER BY part"
        for gameid, _, _ in user_games["aaa"] + user_games["bbb"]:
            for ttyrec in conn.execute(cmd, (gameid,)).fetchall():
                assert ttyrec[1] == 0

        for gameid, _, _ in user_games["ccc"]:
            for i, ttyrec in enumerate(conn.execute(cmd, (gameid,)).fetchall()):
                assert ttyrec[1] == i

    def test_version(self):
        # Expect adding nledata to provide latest ttyrec_version
        ttyrec_version = db.get_ttyrec_version("nletest")
        assert ttyrec_version == nle.nethack.TTYREC_VERSION

        # Expect adding altorg data additions to give version 1
        ttyrec_version = db.get_ttyrec_version("altorgtest")
        assert ttyrec_version == 1

        # Expect basictest using old ttyrecs to give version 1
        ttyrec_version = db.get_ttyrec_version("basictest")
        assert ttyrec_version == 1



import sys

from sample_factory.algorithms.utils.arguments import arg_parser
from sample_factory.algorithms.utils.arguments import parse_args
from sample_factory.run_algorithm import run_algorithm

# Needs to be imported to register models and envs
import nle.language_wrapper.agents.sample_factory.env  # pylint: disable=['unused-import']
import nle.language_wrapper.agents.sample_factory.language_encoder  # pylint: disable=['unused-import']
from nle.language_wrapper.agents.sample_factory.common import custom_parse_args


def parse_all_args(argv=None, evaluation=False):
    parser = arg_parser(argv, evaluation=evaluation)
    cfg = parse_args(argv=argv, evaluation=evaluation, parser=parser)
    return cfg


def main():
    cfg = custom_parse_args()
    status = run_algorithm(cfg)
    return status


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
#
# Copyright (c) Facebook, Inc. and its affiliates.
from nle.nethack import actions
from nle.scripts import ttyplay


def main():
    ttyplay.INPUTS = actions._ACTIONS_DICT
    ttyplay.main()


if __name__ == "__main__":
    main()



import sys

from sample_factory.algorithms.appo.enjoy_appo import enjoy

import nle.language_wrapper.agents.sample_factory.env  # pylint: disable=['unused-import']
import nle.language_wrapper.agents.sample_factory.language_encoder  # pylint: disable=['unused-import']
from nle.language_wrapper.agents.sample_factory.common import custom_parse_args


def main():
    cfg = custom_parse_args(evaluation=True)
    status = enjoy(cfg)
    return status


if __name__ == "__main__":
    sys.exit(main())



from nle.language_wrapper import NLELanguageWrapper
from nle.language_wrapper.tests.test_utils import strs_to_2d


def test_message_spell_menu(fake_nle_env):
    menu = [
        "   Choose which spell to cast                               ",
        "                                                             ",
        "   Name                 Level Category     Fail Retention   ",
        "   a - healing                1   healing        0%      100%",
        "   (end)                                                     ",
    ]
    empty_lines = [
        "                                                      ",
        "                                                      ",
    ]
    screen_map = [
        "         -----------------------------------------|           ",
        "         |               @                        |           ",
    ]
    # Fill with spaces for menu
    tty_chars = strs_to_2d(menu + empty_lines + screen_map, fill_value=32)

    fake_nle_env.reset.return_value["tty_chars"] = tty_chars
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()

    expected_menu = (
        "Choose which spell to cast\n"
        "Name                 Level Category     Fail Retention\n"
        "a - healing                1   healing        0%      100%\n"
        "(end)"
    )
    assert obsv["text_message"] == expected_menu


def test_message_more_end(fake_nle_env):
    things_that_are_here = [
        " Things that are here:      ",
        " a goblin corpse             ",
        " an iron skull cap           ",
        " --More--                    ",
    ]
    screen_map = [
        "         -----------------------------------------|           ",
        "         |               @                        |           ",
    ]
    # Fill with spaces for menu
    tty_chars = strs_to_2d(things_that_are_here + screen_map, fill_value=32)

    fake_nle_env.reset.return_value["tty_chars"] = tty_chars
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()

    expected_menu = (
        "Things that are here:\n"
        + "a goblin corpse\n"
        + "an iron skull cap\n"
        + "--More--"
    )
    assert obsv["text_message"] == expected_menu


def test_message_full_stop_end(fake_nle_env):
    message = [" It's a wall. ", " "]

    screen_map = [
        "         -----------------------------------------|           ",
        "         |               @                        |           ",
    ]

    # Fill with spaces for menu
    tty_chars = strs_to_2d(message + screen_map, fill_value=32)

    fake_nle_env.reset.return_value["tty_chars"] = tty_chars
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()

    expected_menu = "It's a wall."
    assert obsv["text_message"] == expected_menu


def test_message_bracket_end(fake_nle_env):
    message = [" What do you want to drop? [$a-k or ?*] ", " "]

    screen_map = [
        "         -----------------------------------------|           ",
        "         |               @                        |           ",
    ]
    # Fill with spaces for menu
    tty_chars = strs_to_2d(message + screen_map, fill_value=32)
    fake_nle_env.reset.return_value["tty_chars"] = tty_chars
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    expected_menu = "What do you want to drop? [$a-k or ?*]"
    assert obsv["text_message"] == expected_menu


def test_message_parenthesis_end(fake_nle_env):
    message = [" Where do you want to travel to?  (For instructions type a '?') ", " "]

    screen_map = [
        "         -----------------------------------------|           ",
        "         |               @                        |           ",
    ]
    # Fill with spaces for menu
    tty_chars = strs_to_2d(message + screen_map, fill_value=32)
    fake_nle_env.reset.return_value["tty_chars"] = tty_chars
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    expected_message = "Where do you want to travel to?  (For instructions type a '?')"
    assert obsv["text_message"] == expected_message


def test_message_multipage(fake_nle_env):
    message = [
        " Extended Commands List    ",
        "                               ",
        " a - Hide commands that don't autocomplete (those not marked with [A])   ",
        " : - Search extended commands      ",
        " ",
        " Extended commands                     ",
        " #                  perform an extended command  ",
        " ?              [A] list all extended commands  ",
        " adjust         [A] adjust inventory letters   ",
        " annotate       [A] name current level  ",
        " (1 of 5) ",
    ]

    tty_chars = strs_to_2d(message, fill_value=32)
    fake_nle_env.reset.return_value["tty_chars"] = tty_chars
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    expected_message = (
        "Extended Commands List\n"
        "a - Hide commands that don't autocomplete (those not marked with [A])\n"
        ": - Search extended commands\n"
        "Extended commands\n"
        "#                  perform an extended command\n"
        "?              [A] list all extended commands\n"
        "adjust         [A] adjust inventory letters\n"
        "annotate       [A] name current level\n"
        "(1 of 5)"
    )
    assert obsv["text_message"] == expected_message


def test_message_takeoffall(fake_nle_env):
    message = [
        " What type of things do you want to take off?    ",
        " ",
        "  a - All worn types                             ",
        " b - Weapons ",
        " c - Armor ",
        " U - Items known to be Uncursed ",
        " (end) ",
    ]
    screen_map = [
        "         -----------------------------------------|           ",
        "         |               @                        |           ",
    ]
    # Fill with spaces for menu
    tty_chars = strs_to_2d(message + screen_map, fill_value=32)
    fake_nle_env.reset.return_value["tty_chars"] = tty_chars
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    expected_message = (
        "What type of things do you want to take off?\n"
        "a - All worn types\n"
        "b - Weapons\n"
        "c - Armor\n"
        "U - Items known to be Uncursed\n"
        "(end)"
    )
    assert obsv["text_message"] == expected_message


def test_filter_map_from_conduct(fake_nle_env):
    message = [
        "                          Voluntary challenges:                  ",
        "                           You have gone without food.              ",
        "                           You have been an atheist.                ",
        "              -----        You have never hit with a wielded weapon.",
        "             #+....######  You have been a pacifist.                ",
        "             #+....######  --More--                ",
    ]
    tty_chars = strs_to_2d(message, fill_value=32)
    fake_nle_env.reset.return_value["tty_chars"] = tty_chars
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    expected_message = (
        "Voluntary challenges:\n"
        "You have gone without food.\n"
        "You have been an atheist.\n"
        "You have never hit with a wielded weapon.\n"
        "You have been a pacifist.\n"
        "--More--"
    )
    assert obsv["text_message"] == expected_message


def test_empty_tty_chars_returns_empty_message(fake_nle_env):
    message = [" "]
    tty_chars = strs_to_2d(message, fill_value=32)
    fake_nle_env.reset.return_value["tty_chars"] = tty_chars
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    assert obsv["text_message"] == ""


def test_death_message(fake_nle_env):
    message = [
        "" + "No  Points     Name                 Hp [max] \n"
        "\n"
        "0  Agent-Cav-Hum-Mal-Law died in The Dungeons of Doom on            \n"
        "\n"
        "level 1.  Burned by molten lava.                        -  [16]"
    ]
    tty_chars = strs_to_2d(message, fill_value=32)
    fake_nle_env.reset.return_value["tty_chars"] = tty_chars
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    assert obsv["text_message"] == (
        "No  Points     Name                 Hp [max] \n\n"
        "0  Agent-Cav-Hum-Mal-Law died in The Dungeons of Doom on            \n\n"
        "level 1.  Burned by molten lava.                        -  [16]"
    )


def test_filter_map_from_name(fake_nle_env):
    message = [
        "                              What do you want to name? ",
        "                                                        ",
        "---------                     m - a monster ",
        "|........                     i - a particular object in inventory ",
        "|.......|                     o - the type of an object in inventory  ",
        "|....@.%|                     f - the type of an object upon the floor  ",
        "|!...df.|                     d - the type of an object on discoveries list ",
        "|.......+                     a - record an annotation for the current level ",
        "--+------                     (end)",
    ]
    tty_chars = strs_to_2d(message, fill_value=32)
    fake_nle_env.reset.return_value["tty_chars"] = tty_chars
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    expected_message = (
        "What do you want to name?\n"
        "m - a monster\n"
        "i - a particular object in inventory\n"
        "o - the type of an object in inventory\n"
        "f - the type of an object upon the floor\n"
        "d - the type of an object on discoveries list\n"
        "a - record an annotation for the current level\n"
        "(end)"
    )
    assert obsv["text_message"] == expected_message


def test_filter_map_travel(fake_nle_env):
    message = ["doorway      ", "", "      ------------ ", "      |..........| "]
    tty_chars = strs_to_2d(message, fill_value=32)
    fake_nle_env.reset.return_value["tty_chars"] = tty_chars
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    expected_message = "doorway"
    assert obsv["text_message"] == expected_message



from gym import Wrapper
from gym import spaces
from nle.env import NLE
from nle.nethack import actions as nethack_actions

from nle import nle_language_obsv


class NLELanguageWrapper(Wrapper):
    @property
    def spec(self):
        return self.env.spec

    all_nle_action_map = {
        nethack_actions.UnsafeActions.HELP: ["help", "?"],
        nethack_actions.UnsafeActions.PREVMSG: ["previous message", "^p"],
        nethack_actions.CompassDirection.N: ["north", "k"],
        nethack_actions.CompassDirection.E: ["east", "l"],
        nethack_actions.CompassDirection.S: ["south", "j"],
        nethack_actions.CompassDirection.W: ["west", "h"],
        nethack_actions.CompassIntercardinalDirection.NE: ["northeast", "u"],
        nethack_actions.CompassIntercardinalDirection.SE: ["southeast", "n"],
        nethack_actions.CompassIntercardinalDirection.SW: ["southwest", "b"],
        nethack_actions.CompassIntercardinalDirection.NW: ["northwest", "y"],
        nethack_actions.CompassCardinalDirectionLonger.N: ["far north", "K"],
        nethack_actions.CompassCardinalDirectionLonger.E: ["far east", "L"],
        nethack_actions.CompassCardinalDirectionLonger.S: ["far south", "J"],
        nethack_actions.CompassCardinalDirectionLonger.W: ["far west", "H"],
        nethack_actions.CompassIntercardinalDirectionLonger.NE: ["far northeast", "U"],
        nethack_actions.CompassIntercardinalDirectionLonger.SE: ["far southeast", "N"],
        nethack_actions.CompassIntercardinalDirectionLonger.SW: ["far southwest", "B"],
        nethack_actions.CompassIntercardinalDirectionLonger.NW: ["far northwest", "Y"],
        nethack_actions.MiscDirection.UP: ["up", "<"],
        nethack_actions.MiscDirection.DOWN: ["down", ">"],
        nethack_actions.MiscDirection.WAIT: ["wait", "."],
        nethack_actions.MiscAction.MORE: ["more", "\r", r"\r"],
        nethack_actions.Command.EXTCMD: ["extcmd", "#"],
        nethack_actions.Command.EXTLIST: ["extlist", "M-?"],
        nethack_actions.Command.ADJUST: ["adjust", "M-a"],
        nethack_actions.Command.ANNOTATE: ["annotate", "M-A"],
        nethack_actions.Command.APPLY: ["apply", "a"],
        nethack_actions.Command.ATTRIBUTES: ["attributes", "^x"],
        nethack_actions.Command.AUTOPICKUP: ["autopickup", "@"],
        nethack_actions.Command.CALL: ["call", "C"],
        nethack_actions.Command.CAST: ["cast", "Z"],
        nethack_actions.Command.CHAT: ["chat", "M-c"],
        nethack_actions.Command.CLOSE: ["close", "c"],
        nethack_actions.Command.CONDUCT: ["conduct", "M-C"],
        nethack_actions.Command.DIP: ["dip", "M-d"],
        nethack_actions.Command.DROP: ["drop", "d"],
        nethack_actions.Command.DROPTYPE: ["droptype", "D"],
        nethack_actions.Command.EAT: ["eat", "e"],
        nethack_actions.Command.ESC: ["esc", "^["],
        nethack_actions.Command.ENGRAVE: ["engrave", "E"],
        nethack_actions.Command.ENHANCE: ["enhance", "M-e"],
        nethack_actions.Command.FIRE: ["fire", "f"],
        nethack_actions.Command.FIGHT: ["fight", "F"],
        nethack_actions.Command.FORCE: ["force", "M-f"],
        nethack_actions.Command.GLANCE: ["glance", ";"],
        nethack_actions.Command.HISTORY: ["history", "V"],
        nethack_actions.Command.INVENTORY: ["inventory", "i"],
        nethack_actions.Command.INVENTTYPE: ["inventtype", "I"],
        nethack_actions.Command.INVOKE: ["invoke", "M-i"],
        nethack_actions.Command.JUMP: ["jump", "M-j"],
        nethack_actions.Command.KICK: ["kick", "^d"],
        nethack_actions.Command.KNOWN: ["known", "\\"],
        nethack_actions.Command.KNOWNCLASS: ["knownclass", "`"],
        nethack_actions.Command.LOOK: ["look", ":"],
        nethack_actions.Command.LOOT: ["loot", "M-l"],
        nethack_actions.Command.MONSTER: ["monster", "M-m"],
        nethack_actions.Command.MOVE: ["move", "m"],
        nethack_actions.Command.MOVEFAR: ["movefar", "M"],
        nethack_actions.Command.OFFER: ["offer", "M-o"],
        nethack_actions.Command.OPEN: ["open", "o"],
        nethack_actions.Command.OPTIONS: ["options", "O"],
        nethack_actions.Command.OVERVIEW: ["overview", "^o"],
        nethack_actions.Command.PAY: ["pay", "p"],
        nethack_actions.Command.PICKUP: ["pickup", ","],
        nethack_actions.Command.PRAY: ["pray", "M-p"],
        nethack_actions.Command.PUTON: ["puton", "P"],
        nethack_actions.Command.QUAFF: ["quaff", "q"],
        nethack_actions.Command.QUIT: ["quit", "M-q"],
        nethack_actions.Command.QUIVER: ["quiver", "Q"],
        nethack_actions.Command.READ: ["read", "r"],
        nethack_actions.Command.REDRAW: ["redraw", "^r"],
        nethack_actions.Command.REMOVE: ["remove", "R"],
        nethack_actions.Command.RIDE: ["ride", "M-R"],
        nethack_actions.Command.RUB: ["rub", "M-r"],
        nethack_actions.Command.RUSH: ["rush", "g"],
        nethack_actions.Command.RUSH2: ["rush2", "G"],
        nethack_actions.Command.SAVE: ["save", "S"],
        nethack_actions.Command.SEARCH: ["search", "s"],
        nethack_actions.Command.SEEALL: ["seeall", "*"],
        nethack_actions.Command.SEEAMULET: ["seeamulet", '"'],
        nethack_actions.Command.SEEARMOR: ["seearmor", "["],
        nethack_actions.Command.SEEGOLD: ["seegold", "dollar", "$"],
        nethack_actions.Command.SEERINGS: ["seerings", "="],
        nethack_actions.Command.SEESPELLS: ["seespells", "plus", "+"],
        nethack_actions.Command.SEETOOLS: ["seetools", "("],
        nethack_actions.Command.SEETRAP: ["seetrap", "^"],
        nethack_actions.Command.SEEWEAPON: ["seeweapon", ")"],
        nethack_actions.Command.SHELL: ["shell", "!"],
        nethack_actions.Command.SIT: ["sit", "M-s"],
        nethack_actions.Command.SWAP: ["swap", "x"],
        nethack_actions.Command.TAKEOFF: ["takeoff", "T"],
        nethack_actions.Command.TAKEOFFALL: ["takeoffall", "A"],
        nethack_actions.Command.TELEPORT: ["teleport", "^t"],
        nethack_actions.Command.THROW: ["throw", "t"],
        nethack_actions.Command.TIP: ["tip", "M-T"],
        nethack_actions.Command.TRAVEL: ["travel", "_"],
        nethack_actions.Command.TURN: ["turnundead", "M-t"],
        nethack_actions.Command.TWOWEAPON: ["twoweapon", "X"],
        nethack_actions.Command.UNTRAP: ["untrap", "M-u"],
        nethack_actions.Command.VERSION: ["version", "M-v"],
        nethack_actions.Command.VERSIONSHORT: ["versionshort", "v"],
        nethack_actions.Command.WEAR: ["wear", "W"],
        nethack_actions.Command.WHATDOES: ["whatdoes", "&"],
        nethack_actions.Command.WHATIS: ["whatis", "/"],
        nethack_actions.Command.WIELD: ["wield", "w"],
        nethack_actions.Command.WIPE: ["wipe", "M-w"],
        nethack_actions.Command.ZAP: ["zap", "z"],
        nethack_actions.TextCharacters.MINUS: ["minus", "-"],
        nethack_actions.TextCharacters.SPACE: ["space", " "],
        nethack_actions.TextCharacters.APOS: ["apos", "'"],
        nethack_actions.TextCharacters.NUM_0: ["zero", "0"],
        nethack_actions.TextCharacters.NUM_1: ["one", "1"],
        nethack_actions.TextCharacters.NUM_2: ["two", "2"],
        nethack_actions.TextCharacters.NUM_3: ["three", "3"],
        nethack_actions.TextCharacters.NUM_4: ["four", "4"],
        nethack_actions.TextCharacters.NUM_5: ["five", "5"],
        nethack_actions.TextCharacters.NUM_6: ["six", "6"],
        nethack_actions.TextCharacters.NUM_7: ["seven", "7"],
        nethack_actions.TextCharacters.NUM_8: ["eight", "8"],
        nethack_actions.TextCharacters.NUM_9: ["nine", "9"],
        nethack_actions.WizardCommand.WIZDETECT: ["wizard detect", "^e"],
        nethack_actions.WizardCommand.WIZGENESIS: ["wizard genesis", "^g"],
        nethack_actions.WizardCommand.WIZIDENTIFY: ["wizard identify", "^i"],
        nethack_actions.WizardCommand.WIZLEVELPORT: ["wizard teleport", "^v"],
        nethack_actions.WizardCommand.WIZMAP: ["wizard map", "^f"],
        nethack_actions.WizardCommand.WIZWHERE: ["wizard where", "^o"],
        nethack_actions.WizardCommand.WIZWISH: ["wizard wish", "^w"],
    }

    REQUIRED_NLE_OBSV_KEYS = {
        "glyphs",
        "blstats",
        "tty_cursor",
        "inv_strs",
        "inv_letters",
        "tty_chars",
    }

    def step(self, action):
        if self.use_language_action:
            action = self.pre_step(action)
        nle_obsv, reward, done, info = self.env.step(action)
        return self.post_step(nle_obsv), reward, done, info

    def pre_step(self, action: str):
        """Translate language action to nle action.
        Args:
            action (str): language/text action
        Returns:
            (Enum): nle action
        """
        if action not in self.action_str_enum_map:
            raise ValueError(
                f"Action {repr(action)} is not recognized "
                "or not supported for this environment"
            )
        nle_action_enum = self.action_str_enum_map[action]
        nle_action_idx = self.action_enum_index_map[nle_action_enum]
        return nle_action_idx

    def nle_obsv_to_language(self, nle_obsv):
        """Translate NLE Observation into a language observation.
        Args:
            nle_obsv (dict): NLE observation from the base environment
        Returns:
            (dict): language observation
        """
        glyphs = nle_obsv["glyphs"]
        blstats = nle_obsv["blstats"]
        tty_cursor = nle_obsv["tty_cursor"]
        inv_strs = nle_obsv["inv_strs"]
        inv_letters = nle_obsv["inv_letters"]
        tty_chars = nle_obsv["tty_chars"]
        return {
            "text_glyphs": self.nle_language.text_glyphs(glyphs, blstats).decode(
                "latin-1"
            ),
            "text_message": self.nle_language.text_message(tty_chars).decode("latin-1"),
            "text_blstats": self.nle_language.text_blstats(blstats).decode("latin-1"),
            "text_inventory": self.nle_language.text_inventory(
                inv_strs, inv_letters
            ).decode("latin-1"),
            "text_cursor": self.nle_language.text_cursor(
                glyphs, blstats, tty_cursor
            ).decode("latin-1"),
        }

    def post_step(self, nle_obsv):
        """Post step operations. Used for translating the observation
        Args:
            nle_obsv (dict): nle observation from base environment
        Returns:
            (dict): language observation
        """
        return self.nle_obsv_to_language(nle_obsv)

    def reset(self, **kwargs):
        self.pre_reset()
        obsv = self.env.reset(**kwargs)
        return self.post_reset(obsv)

    def pre_reset(self):
        """Pre reset operations."""

    def post_reset(self, obsv):
        """Post reset operations.  Translate the NLE observation into language version
        Args:
            nle_obsv (dict): nle observation from base environment
        Returns:
            (dict): language observation
        """
        return self.nle_obsv_to_language(obsv)

    def __init__(self, env, use_language_action=True):
        """Initialize the wrapper
        Args:
            env (nle.env.NLE): NLE based environment to be wrapped
            use_language_action(bool): Use language action or discrete integer actions
        """
        super().__init__(env)
        assert isinstance(env.unwrapped, NLE), "Only NLE environments are supported"
        missing_obsv_keys = self.REQUIRED_NLE_OBSV_KEYS.difference(
            env.observation_space.spaces.keys()
        )
        assert (
            len(missing_obsv_keys) == 0
        ), f"NLE environment missing required obsv key(s): {missing_obsv_keys}"
        # assert observations are included
        self.use_language_action = use_language_action
        self.nle_language = nle_language_obsv.NLELanguageObsv()

        # Build map for action string to NLE Action Enum
        self.action_str_enum_map = {}
        for nle_action_enum, action_strs in self.all_nle_action_map.items():
            if nle_action_enum in self.env.actions:
                for action_str in action_strs:
                    self.action_str_enum_map[action_str] = nle_action_enum

        # Build map for NLE Action Enum to NLE action index
        self.action_enum_index_map = {}
        for nle_action_enum, _ in self.all_nle_action_map.items():
            if nle_action_enum in self.env.actions:
                self.action_enum_index_map[nle_action_enum] = self.env.actions.index(
                    nle_action_enum
                )

        if self.use_language_action:
            self.action_space = spaces.Space()
        self.observation_space = spaces.Dict(
            {
                "text_glyphs": spaces.Space(),
                "text_message": spaces.Space(),
                "text_blstats": spaces.Space(),
                "text_inventory": spaces.Space(),
                "text_cursor": spaces.Space(),
            }
        )



import gym
import numpy as np
import pytest
from gym import spaces
from nle.env import NLE
from nle.nethack import actions as nethack_actions

from nle.language_wrapper.tests.test_utils import str_to_1d
from nle.language_wrapper.tests.test_utils import strs_to_2d


@pytest.fixture
def real_nethack_env():
    """Fixture for real NLE environment"""
    nle_env = gym.make("NetHackChallenge-v0")
    return nle_env


@pytest.fixture
def fake_nle_env(mocker):
    """Fixture for fake NLE environment"""
    nle_env = mocker.MagicMock(spec=NLE)
    blstats = np.array(
        [
            66,  # player Y
            14,  # Player X
            19,  # Percentile strength
            1,  # Strength
            4,  # Dexterity
            5,  # Constitution
            6,  # Intelligence
            7,  # Wisdom
            8,  # Charisma
            123,  # Score
            11,  # HP Current
            12,  # HP Max
            9,  # Depth
            10,  # Gold
            1,  # Energy current
            13,  # Energy max
            1,  # AC
            23,  # Monster level
            7,  # Xp Current
            10,  # XP level
            5,  # Time
            4,  # Hunger
            2,  # Encumberance
            21,  # Dungeon number
            10,  # Level number
            11,  # Conditions
            -1,  # Alignment
        ],
        dtype=np.int64,
    )
    tty_cursor = np.array(
        [14 + 1, 66],  # Player X (offset by 1 for some reason)  # Player Y
        dtype=np.uint8,
    )
    tty_chars = strs_to_2d(
        ["Hello Agent, welcome to NetHack!  You are a neutral gnomish Caveman."]
    )
    inv_letters = str_to_1d("ab\x00")
    inv_strs = strs_to_2d(
        [
            "a blessed +1 quarterstaff (weapon in hands)\x00",
            "an uncursed +0 cloak of magic resistance (being worn)\x00",
        ]
    )
    glyphs = np.full((21, 79), 2359)
    glyphs[14, 66] = 333
    glyphs[20, 50] = 397  # tame dog
    glyphs[10, 65] = 2360
    glyphs[9, 64] = 2390  # fountain
    glyphs[13, 66] = 2374  # vertical closed door
    obsv = {
        "glyphs": glyphs,
        "blstats": blstats,
        "tty_chars": tty_chars,
        "inv_letters": inv_letters,
        "inv_strs": inv_strs,
        "tty_cursor": tty_cursor,
    }
    reward = 0.0
    done = False
    info = None
    nle_env.reset = mocker.MagicMock(return_value=obsv)
    nle_env.step = mocker.MagicMock(return_value=(obsv, reward, done, info))
    nle_env.actions = [nethack_actions.CompassDirection.N]
    nle_env.observation_space = spaces.Dict(
        {
            "glyphs": spaces.Space(),
            "blstats": spaces.Space(),
            "tty_cursor": spaces.Space(),
            "inv_strs": spaces.Space(),
            "inv_letters": spaces.Space(),
            "tty_chars": spaces.Space(),
        }
    )
    return nle_env


@pytest.fixture
def fake_nethack_multiple_monsters_env(mocker):
    """Fixture for fake NLE environment containing multiple monsters"""
    nle_env = mocker.MagicMock(spec=NLE)
    blstats = np.array([66, 14, 19], dtype=np.uint8)
    tty_cursor = np.array([66, 14, 19], dtype=np.uint8)
    tty_chars = strs_to_2d(
        ["Hello Agent, welcome to NetHack!  You are a neutral gnomish Caveman."]
    )
    inv_letters = str_to_1d("ab\x00")

    inv_strs = strs_to_2d(
        [
            "a blessed +1 quarterstaff (weapon in hands)\x00",
            "an uncursed +0 cloak of magic resistance (being worn)\x00",
        ]
    )
    glyphs = np.full((21, 79), 2359)
    glyphs[14, 66] = 359
    glyphs[14, 18] = 397  # tame dog 1
    glyphs[20, 50] = 397  # tame dog 2
    glyphs[20, 51] = 397  # tame dog 3
    glyphs[10, 65] = 2360
    glyphs[9, 64] = 2390  # fountain
    glyphs[9, 63] = 2390  # fountain
    glyphs[13, 66] = 2374  # vertical closed door
    obsv = {
        "glyphs": glyphs,
        "blstats": blstats,
        "tty_chars": tty_chars,
        "inv_letters": inv_letters,
        "inv_strs": inv_strs,
        "tty_cursor": tty_cursor,
    }
    reward = 0.0
    done = False
    info = None
    nle_env.reset = mocker.MagicMock(return_value=obsv)
    nle_env.step = mocker.MagicMock(return_value=(obsv, reward, done, info))
    nle_env.actions = [nethack_actions.CompassDirection.N]
    nle_env.observation_space = spaces.Dict(
        {
            "glyphs": spaces.Space(),
            "blstats": spaces.Space(),
            "tty_cursor": spaces.Space(),
            "inv_strs": spaces.Space(),
            "inv_letters": spaces.Space(),
            "tty_chars": spaces.Space(),
        }
    )
    return nle_env



# Copyright (c) Facebook, Inc. and its affiliates.
import os
import shutil
import sys
import tempfile
import warnings
import weakref

import numpy as np
import pkg_resources

from nle import _pynethack

DLPATH = os.path.join(os.path.dirname(_pynethack.__file__), "libnethack.so")

DUNGEON_SHAPE = (_pynethack.nethack.ROWNO, _pynethack.nethack.COLNO - 1)
BLSTATS_SHAPE = (_pynethack.nethack.NLE_BLSTATS_SIZE,)
MESSAGE_SHAPE = (_pynethack.nethack.NLE_MESSAGE_SIZE,)
PROGRAM_STATE_SHAPE = (_pynethack.nethack.NLE_PROGRAM_STATE_SIZE,)
INTERNAL_SHAPE = (_pynethack.nethack.NLE_INTERNAL_SIZE,)
MISC_SHAPE = (_pynethack.nethack.NLE_MISC_SIZE,)
INV_SIZE = (_pynethack.nethack.NLE_INVENTORY_SIZE,)
INV_STRS_SHAPE = (
    _pynethack.nethack.NLE_INVENTORY_SIZE,
    _pynethack.nethack.NLE_INVENTORY_STR_LENGTH,
)
SCREEN_DESCRIPTIONS_SHAPE = DUNGEON_SHAPE + (
    _pynethack.nethack.NLE_SCREEN_DESCRIPTION_LENGTH,
)
TERMINAL_SHAPE = (_pynethack.nethack.NLE_TERM_LI, _pynethack.nethack.NLE_TERM_CO)

OBSERVATION_DESC = {
    "glyphs": dict(shape=DUNGEON_SHAPE, dtype=np.int16),
    "chars": dict(shape=DUNGEON_SHAPE, dtype=np.uint8),
    "colors": dict(shape=DUNGEON_SHAPE, dtype=np.uint8),
    "specials": dict(shape=DUNGEON_SHAPE, dtype=np.uint8),
    "blstats": dict(shape=BLSTATS_SHAPE, dtype=np.int64),
    "message": dict(shape=MESSAGE_SHAPE, dtype=np.uint8),
    "program_state": dict(shape=PROGRAM_STATE_SHAPE, dtype=np.int32),
    "internal": dict(shape=INTERNAL_SHAPE, dtype=np.int32),
    "inv_glyphs": dict(shape=INV_SIZE, dtype=np.int16),
    "inv_letters": dict(shape=INV_SIZE, dtype=np.uint8),
    "inv_oclasses": dict(shape=INV_SIZE, dtype=np.uint8),
    "inv_strs": dict(shape=INV_STRS_SHAPE, dtype=np.uint8),
    "screen_descriptions": dict(shape=SCREEN_DESCRIPTIONS_SHAPE, dtype=np.uint8),
    "tty_chars": dict(shape=TERMINAL_SHAPE, dtype=np.uint8),
    "tty_colors": dict(shape=TERMINAL_SHAPE, dtype=np.int8),
    "tty_cursor": dict(shape=(2,), dtype=np.uint8),
    "misc": dict(shape=MISC_SHAPE, dtype=np.int32),
}


NETHACKOPTIONS = (
    "autopickup",
    "color",
    "disclose:+i +a +v +g +c +o",
    "mention_walls",
    "nobones",
    "nocmdassist",
    "nolegacy",
    "nosparkle",
    "pickup_burden:unencumbered",
    "pickup_types:$?!/",
    "runmode:teleport",
    "showexp",
    "showscore",
    "time",
)

HACKDIR = pkg_resources.resource_filename("nle", "nethackdir")
TTYREC_VERSION = 3


def _new_dl_linux(vardir):
    if hasattr(os, "memfd_create"):
        target = os.memfd_create("nle.so")
        path = "/proc/self/fd/%i" % target
        try:
            shutil.copyfile(DLPATH, path)  # Should use sendfile.
        except IOError:
            os.close(target)
            raise
        return os.fdopen(target), path

    # Otherwise, no memfd_create. Try with O_TMPFILE via the tempfile module.
    dl = tempfile.TemporaryFile(suffix="libnethack.so", dir=vardir)
    path = "/proc/self/fd/%i" % dl.fileno()
    shutil.copyfile(DLPATH, path)  # Should use sendfile.
    return dl, path


def _new_dl(vardir):
    """Creates a copied .so file to allow for multiple independent NLE instances"""
    if sys.platform == "linux":
        return _new_dl_linux(vardir)

    # MacOS has no memfd_create or O_TMPFILE. Using /dev/fd/{FD} as an argument
    # to dlopen doesn't work after unlinking from the file system. So let's copy
    # instead and hope vardir gets properly deleted at some point.
    dl = tempfile.NamedTemporaryFile(suffix="libnethack.so", dir=vardir)
    shutil.copyfile(DLPATH, dl.name)  # Might use fcopyfile.
    return dl, dl.name


def _close(pynethack, dl, tempdir, warn=True):
    if pynethack is not None:
        pynethack.close()
    if dl is not None:
        dl.close()
    if tempdir is not None:
        tempdir.cleanup()
    if warn:
        warnings.warn("nethack.Nethack instance not closed", ResourceWarning)


def tty_render(chars, colors, cursor=None):
    """Returns chars as string with ANSI escape sequences.

    Args:
      chars: A row x columns numpy array of chars.
      colors: A numpy array of colors (0-15), same shape as chars.
      cursor: An optional (row, column) index for the cursor,
        displayed as underlined.

    Returns:
      A string with chars decorated by ANSI escape sequences.
    """
    rows, cols = chars.shape
    if cursor is None:
        cursor = (-1, -1)
    cursor = tuple(cursor)
    result = ""
    for i in range(rows):
        result += "\n"
        for j in range(cols):
            entry = "\033[%d;3%dm%s" % (
                # & 8 checks for brightness.
                bool(colors[i, j] & 8),
                colors[i, j] & ~8,
                chr(chars[i, j]),
            )
            if cursor != (i, j):
                result += entry
            else:
                result += "\033[4m%s\033[0m" % entry
    return result + "\033[0m"


# TODO: Not thread-safe for many reasons.
class Nethack:
    _instances = 0

    def __init__(
        self,
        observation_keys=OBSERVATION_DESC.keys(),
        playername="Agent-mon-hum-neu-mal",
        ttyrec="nle.ttyrec%i.bz2" % TTYREC_VERSION,
        options=None,
        copy=False,
        wizard=False,
        hackdir=HACKDIR,
        spawn_monsters=True,
        scoreprefix="",
    ):
        self._copy = copy

        if not os.path.exists(hackdir) or not os.path.exists(
            os.path.join(hackdir, "nhdat")
        ):
            raise FileNotFoundError(
                "Couldn't find NetHack installation at '%s'." % hackdir
            )

        # Create a HACKDIR for us.
        self._tempdir = tempfile.TemporaryDirectory(prefix="nle")
        self._vardir = self._tempdir.name

        # Symlink a nhdat.
        os.symlink(os.path.join(hackdir, "nhdat"), os.path.join(self._vardir, "nhdat"))

        # Touch files, so lock_file() in files.c passes.
        for fn in ["perm", "record", "logfile"]:
            os.close(os.open(os.path.join(self._vardir, fn), os.O_CREAT))
        if scoreprefix:
            os.close(os.open(scoreprefix + "xlogfile", os.O_CREAT))
        else:
            os.close(os.open(os.path.join(self._vardir, "xlogfile"), os.O_CREAT))

        os.mkdir(os.path.join(self._vardir, "save"))

        # An assortment of hacks:
        #   Copy our .so into self._vardir to load several copies of the dl.
        #   (Or use a memfd_create hack to create a file that gets deleted on
        #    process exit.)
        self._dl, self.dlpath = _new_dl(self._vardir)

        # Finalize even when the rest of this constructor fails.
        self._finalizer = weakref.finalize(self, _close, None, self._dl, self._tempdir)

        if options is None:
            options = NETHACKOPTIONS
        self.options = list(options) + ["name:" + playername]
        if playername.split("-", 1)[1:] == ["@"]:
            # Random role. Unless otherwise specified, randomize
            # race/gender/alignment too.
            for key in ("race", "gender", "align"):
                if not any(o for o in options if o.startswith(key + ":")):
                    self.options.append("%s:random" % key)

        if wizard:
            self.options.append("playmode:debug")
        self._wizard = wizard
        self._nethackoptions = ",".join(self.options)
        if ttyrec is None:
            self._pynethack = _pynethack.Nethack(
                self.dlpath, self._vardir, self._nethackoptions, spawn_monsters
            )
        else:
            self._pynethack = _pynethack.Nethack(
                self.dlpath,
                ttyrec,
                self._vardir,
                self._nethackoptions,
                spawn_monsters,
                scoreprefix,
            )
        self._ttyrec = ttyrec

        self._finalizer.detach()
        self._finalizer = weakref.finalize(
            self, _close, self._pynethack, self._dl, self._tempdir
        )

        self._obs_buffers = {}

        for key in observation_keys:
            if key not in OBSERVATION_DESC:
                raise ValueError("Unknown observation '%s'" % key)
            self._obs_buffers[key] = np.zeros(**OBSERVATION_DESC[key])

        self._pynethack.set_buffers(**self._obs_buffers)

        self._obs = tuple(self._obs_buffers[key] for key in observation_keys)
        if self._copy:
            self._step_return = lambda: tuple(o.copy() for o in self._obs)
        else:
            self._step_return = lambda: self._obs

    def step(self, action):
        self._pynethack.step(action)
        return self._step_return(), self._pynethack.done()

    def reset(self, new_ttyrec=None, wizkit_items=None):
        if wizkit_items is not None:
            if not self._wizard:
                raise ValueError("Set wizard=True to use the wizkit option.")
            # TODO ideally we need to check the validity of the requested items
            self._pynethack.set_wizkit("\n".join(wizkit_items))
        if new_ttyrec is None:
            self._pynethack.reset()
        else:
            self._pynethack.reset(new_ttyrec)
            self._ttyrec = new_ttyrec
        # No seeding performed here: If we fixed the seeds, we'd only
        # get one episode.
        return self._step_return()

    def close(self):
        if self._finalizer.detach():
            _close(self._pynethack, self._dl, self._tempdir, warn=False)
        self._pynethack = None
        self._dl = None
        self._tempdir = None

    def set_initial_seeds(self, core, disp, reseed=False):
        self._pynethack.set_initial_seeds(core, disp, reseed)

    def set_current_seeds(self, core=None, disp=None, reseed=False):
        """Sets the seeds of NetHack right now.

        If either of the three arguments is None, its current value will be
        used instead. Calling with default arguments will not change
        the RNG seeds but disable reseeding. Calling
        `set_current_seeds(reseed=None)` is a no-op and only returns the current
        values. If NetHack detects a good source of random numbers (true on most
        modern systems), its default value for `reseed` before the first call to
        this method is `True`.

        Arguments:
            core [int or None]: Seed for the core RNG.
            disp [int or None]: Seed for the disp (anti-TAS) RNG.
            reseed [boolean or None]: As an Anti-TAS (automation) measure,
                NetHack 3.6 reseeds with true randomness every now and then. This
                flag enables or disables this behavior. If set to True, trajectories
                won't be reproducible.

        Returns:
            [list] the seeds used by NetHack.
        """
        seeds = [core, disp, reseed]
        if any(s is None for s in seeds):
            if all(s is None for s in seeds):
                return
            for i, (s, s0) in enumerate(zip(seeds, self.get_current_seeds())):
                if s is None:
                    seeds[i] = s0
            return self._pynethack.set_seeds(*seeds)
        return self._pynethack.set_seeds(core, disp, reseed)

    def get_current_seeds(self):
        return self._pynethack.get_seeds()

    def in_normal_game(self):
        return self._pynethack.in_normal_game()

    def how_done(self):
        return self._pynethack.how_done()



#!/usr/bin/env python
#
# Copyright (c) Facebook, Inc. and its affiliates.

# Requires
#   pip install pytest-benchmark
# to run
import os

import gym
import numpy as np
import pytest

import nle  # noqa: F401

BASE_KEYS = ["glyphs", "message", "blstats"]
MAPPED_GLYPH = ["chars", "colors", "specials"]
INV_GLYPH = ["inv_glyphs", "inv_strs", "inv_letters", "inv_oclasses"]
SCREEN_DESC = ["screen_descriptions"]
TTY = ["tty_chars", "tty_colors", "tty_cursor"]

EXPERIMENTS = {
    "(1): glyphs/msg/blstats": BASE_KEYS,
    "(2): (1)... + char/col/spec": BASE_KEYS + MAPPED_GLYPH,
    "(3): (2)... + inv_*": BASE_KEYS + MAPPED_GLYPH + INV_GLYPH,
    "(4): (3)... + tty_*": BASE_KEYS + MAPPED_GLYPH + INV_GLYPH + TTY,
    "(5): (4)... + screen_desc": BASE_KEYS
    + MAPPED_GLYPH
    + INV_GLYPH
    + TTY
    + SCREEN_DESC,
}


@pytest.mark.parametrize(
    "observation_keys", EXPERIMENTS.values(), ids=EXPERIMENTS.keys()
)
class TestProfile:
    @pytest.fixture(autouse=True)
    def make_cwd_tmp(self, tmpdir):
        """Makes cwd point to the test's tmpdir."""
        with tmpdir.as_cwd():
            yield

    @pytest.mark.benchmark(disable_gc=True, warmup=False)
    def test_run_1k_steps(self, observation_keys, benchmark):
        if os.environ.get("CI") == "true":
            pytest.skip("Not running benchmark on CI")

        env = gym.make("NetHack-v0", observation_keys=observation_keys)
        seeds = 123456
        steps = 1000

        np.random.seed(seeds)
        actions = np.random.choice(env.action_space.n, size=steps)

        def seed():
            if not nle.nethack.NLE_ALLOW_SEEDING:
                return
            nonlocal seeds
            seeds += 1
            env.seed(seeds, 2 * seeds)

        def play_1k_steps():
            env.reset()
            for a in actions:
                _, _, done, _ = env.step(a)
                if done:
                    seed()
                    env.reset()

        benchmark.pedantic(play_1k_steps, setup=seed, rounds=100, warmup_rounds=10)



# Copyright (c) Facebook, Inc. and its affiliates.
import os
import random
import timeit
import warnings

import numpy as np
import pytest

from nle import _pynethack
from nle import nethack

# MORE + compass directions + long compass directions.
ACTIONS = [
    13,
    107,
    108,
    106,
    104,
    117,
    110,
    98,
    121,
    75,
    76,
    74,
    72,
    85,
    78,
    66,
    89,
]


class TestNetHack:
    @pytest.fixture
    def game(self):  # Make sure we close even on test failure.
        g = nethack.Nethack(observation_keys=("chars", "blstats"))
        try:
            yield g
        finally:
            g.close()

    def test_close_and_restart(self):
        game = nethack.Nethack()
        game.reset()
        game.close()

        game = nethack.Nethack()
        game.reset()
        game.close()

    def test_run_n_episodes(self, tmpdir, game, episodes=3):
        olddir = tmpdir.chdir()  # tmpdir is a py.path.local object.

        chars, blstats = game.reset()

        assert chars.shape == nethack.DUNGEON_SHAPE
        assert blstats.shape == nethack.BLSTATS_SHAPE

        game.step(ord("y"))
        game.step(ord("y"))
        game.step(ord("\n"))

        steps = 0
        start_time = timeit.default_timer()
        start_steps = steps

        mean_sps = 0
        sps_n = 0

        for episode in range(episodes):
            while True:
                ch = random.choice(ACTIONS)
                _, done = game.step(ch)
                if done:
                    # This will typically be DIED, but could be POISONED, etc.
                    assert int(game.how_done()) < int(nethack.GENOCIDED)
                    break

                steps += 1

                if steps % 1000 == 0:
                    end_time = timeit.default_timer()
                    sps = (steps - start_steps) / (end_time - start_time)
                    sps_n += 1
                    mean_sps += (sps - mean_sps) / sps_n
                    print("%f SPS" % sps)
                    start_time = end_time
                    start_steps = steps
            print("Finished episode %i after %i steps." % (episode + 1, steps))
            game.reset()

        print("Finished after %i steps. Mean sps: %f" % (steps, mean_sps))

        # Resetting the game shouldn't have caused cwd to change.
        assert tmpdir.samefile(os.getcwd())

        if mean_sps < 15000:
            warnings.warn("Mean sps was only %f" % mean_sps)
        olddir.chdir()
        # No call to game.close() as fixture will do that for us.

    def test_several_nethacks(self, game):
        game.reset()
        game1 = nethack.Nethack()
        game1.reset()

        try:
            for _ in range(300):
                ch = random.choice(ACTIONS)
                _, done = game.step(ch)
                if done:
                    game.reset()
                _, done = game1.step(ch)
                if done:
                    game1.reset()
        finally:
            game1.close()

    def test_set_initial_seeds(self):
        if not nethack.NLE_ALLOW_SEEDING:
            return  # Nothing to test.
        game = nethack.Nethack(copy=True)
        game.set_initial_seeds(core=42, disp=666)
        obs0 = game.reset()
        try:
            seeds0 = game.get_current_seeds()
            game.set_initial_seeds(core=42, disp=666)
            obs1 = game.reset()
            seeds1 = game.get_current_seeds()
            np.testing.assert_equal(obs0, obs1)
            assert seeds0 == seeds1
        finally:
            game.close()

    def test_set_seed_after_reset(self, game):
        if not nethack.NLE_ALLOW_SEEDING:
            return  # Nothing to test.

        game.reset()
        # Could fail on a system without a good source of randomness:
        assert game.get_current_seeds()[2] is True
        game.set_current_seeds(core=42, disp=666)
        assert game.get_current_seeds() == (42, 666, False)


class TestNetHackFurther:
    def test_run(self):
        game = nethack.Nethack(
            observation_keys=("glyphs", "chars", "colors", "blstats", "program_state")
        )
        _, _, _, _, program_state = game.reset()
        actions = [
            nethack.MiscAction.MORE,
            nethack.MiscAction.MORE,
            nethack.MiscAction.MORE,
            nethack.MiscAction.MORE,
            nethack.MiscAction.MORE,
            nethack.MiscAction.MORE,
        ]

        for action in actions:
            while not program_state[3]:  # in_moveloop.
                obs, done = game.step(nethack.MiscAction.MORE)
                _, _, _, _, program_state = obs

            obs, done = game.step(action)
            if done:
                # Only the good die young.
                obs = game.reset()

            glyphs, chars, colors, blstats, _ = obs

            x, y = blstats[:2]

            assert np.count_nonzero(chars == ord("@")) == 1

            # That's where you're @.
            assert chars[y, x] == ord("@")

            # You're bright (4th bit, 8) white (7), too.
            assert colors[y, x] == 8 ^ 7

            mon = nethack.permonst(nethack.glyph_to_mon(glyphs[y][x]))
            assert mon.mname == "monk"
            assert mon.mlevel == 10

            class_sym = nethack.class_sym.from_mlet(mon.mlet)
            assert class_sym.sym == "@"
            assert class_sym.explain == "human or elf"

        game.close()
        assert os.path.isfile(
            os.path.join(os.getcwd(), "nle.ttyrec%i.bz2" % nethack.TTYREC_VERSION)
        )

    def test_illegal_filename(self):
        with pytest.raises(IOError):
            nethack.Nethack(ttyrec="")
        game = nethack.Nethack()
        with pytest.raises(IOError):
            game.reset("")

    def test_set_buffers_after_reset(self):
        game = nethack.Nethack()
        game.reset()
        with pytest.raises(RuntimeError, match=r"set_buffers called after reset()"):
            game._pynethack.set_buffers()

    def test_nethack_random_character(self):
        game = nethack.Nethack(playername="Hugo-@")
        assert "race:random" in game.options
        assert "gender:random" in game.options
        assert "align:random" in game.options

        game = nethack.Nethack(playername="Jurgen-wiz-gno-cha-mal")
        assert "race:random" not in game.options
        assert "gender:random" not in game.options
        assert "align:random" not in game.options

        game = nethack.Nethack(
            playername="Albert-@",
            options=list(nethack.NETHACKOPTIONS) + ["align:lawful"],
        )
        assert "race:random" in game.options
        assert "gender:random" in game.options
        assert "align:random" not in game.options
        assert "align:lawful" in game.options

        game = nethack.Nethack(
            playername="Rachel",
            options=list(nethack.NETHACKOPTIONS) + ["gender:female"],
        )
        assert "race:random" not in game.options
        assert "gender:random" not in game.options
        assert "align:random" not in game.options
        assert "gender:female" in game.options


class TestNethackSomeObs:
    @pytest.fixture
    def game(self):  # Make sure we close even on test failure.
        g = nethack.Nethack(observation_keys=("program_state", "message", "internal"))
        try:
            yield g
        finally:
            g.close()

    def test_message(self, game):
        messages = []

        program_state, message, _ = game.reset()
        messages.append(message)
        while not program_state[3]:  # in_moveloop.
            (program_state, message, _), done = game.step(nethack.MiscAction.MORE)
            messages.append(message)

        greeting = (
            b"Hello Agent, welcome to NetHack!  You are a neutral male human Monk."
        )
        saw_greeting = True
        for message in messages:
            # `greeting` is often the last message, but not always -- e.g.,
            # it could also be "Be careful!  New moon tonight.".
            assert len(message) == 256
            if (
                memoryview(message)[: len(greeting)] == greeting
                and memoryview(message)[len(greeting)] == 0
            ):
                saw_greeting = True
        assert saw_greeting

    def test_internal(self, game):
        program_state, _, internal = game.reset()
        while not program_state[3]:  # in_moveloop.
            # We are not in moveloop. We might still be in "normal game"
            # if something_worth_saving is true. Example: Startup with
            # "--More--", "Be careful!  New moon tonight."
            assert game.in_normal_game() == program_state[5]  # something_worth_saving.
            (program_state, _, internal), done = game.step(nethack.MiscAction.MORE)

        assert game.in_normal_game()
        assert internal[0] == 1  # deepest_lev_reached.

        (_, _, internal), done = game.step(nethack.Command.INVENTORY)
        assert internal[3] == 1  # xwaitforspace


def get_object(name):
    for index in range(nethack.NUM_OBJECTS):
        obj = nethack.objclass(index)
        if nethack.OBJ_NAME(obj) == name:
            return obj
    else:
        raise ValueError("'%s' not found!" % name)


class TestNethackFunctionsAndConstants:
    def test_permonst_and_class_sym(self):
        glyph = 155  # Lichen.

        mon = nethack.permonst(nethack.glyph_to_mon(glyph))

        assert mon.mname == "lichen"

        cs = nethack.class_sym.from_mlet(mon.mlet)

        assert cs.sym == "F"
        assert cs.explain == "fungus or mold"

        assert nethack.NHW_MESSAGE == 1
        assert hasattr(nethack, "MAXWIN")

        # Slightly irritating to need `chr` here.
        cs = nethack.class_sym.from_oc_class(chr(nethack.WAND_CLASS))
        assert cs.sym == "/"
        assert cs.explain == "wand"

        obj = nethack.objclass(0)
        cs = nethack.class_sym.from_oc_class(obj.oc_class)
        assert cs.sym == "]"
        assert cs.explain == "strange object"

    def test_permonst(self):
        mon = nethack.permonst(0)
        assert mon.mname == "giant ant"
        del mon

        mon = nethack.permonst(1)
        assert mon.mname == "killer bee"

    def test_some_constants(self):
        assert nethack.GLYPH_MON_OFF == 0
        assert nethack.NUMMONS > 300

    def test_illegal_numbers(self):
        with pytest.raises(
            IndexError,
            match=r"should be between 0 and NUMMONS \(%i\) but got %i"
            % (nethack.NUMMONS, nethack.NUMMONS),
        ):
            nethack.permonst(nethack.NUMMONS)

        with pytest.raises(
            IndexError,
            match=r"should be between 0 and NUMMONS \(%i\) but got %i"
            % (nethack.NUMMONS, -1),
        ):
            nethack.permonst(-1)

        with pytest.raises(
            IndexError,
            match=r"should be between 0 and MAXMCLASSES \(%i\) but got 127"
            % nethack.MAXMCLASSES,
        ):
            nethack.class_sym.from_mlet("\x7F")

    def test_objclass(self):
        obj = nethack.objclass(0)
        assert nethack.OBJ_NAME(obj) == "strange object"

        food_ration = get_object("food ration")
        assert food_ration.oc_weight == 20

        elven_dagger = get_object("elven dagger")
        assert nethack.OBJ_DESCR(elven_dagger) == "runed dagger"

    def test_objdescr(self):
        od = nethack.objdescr.from_idx(0)
        assert od.oc_name == "strange object"
        assert od.oc_descr is None

        elven_dagger = get_object("elven dagger")
        od = nethack.objdescr.from_idx(elven_dagger.oc_name_idx)
        assert od.oc_name == "elven dagger"
        assert od.oc_descr == "runed dagger"

        # Example of how to do this with glyphs.
        glyph = nethack.GLYPH_OBJ_OFF + elven_dagger.oc_name_idx
        idx = nethack.glyph_to_obj(glyph)
        assert idx == elven_dagger.oc_name_idx
        assert nethack.objdescr.from_idx(idx) is od

    def test_symdef(self):
        tree_cmap_offset = 18  # Cf. drawing.c.
        tree_glyph = nethack.GLYPH_CMAP_OFF + tree_cmap_offset
        tree_symdef = nethack.symdef.from_idx(tree_cmap_offset)

        assert tree_cmap_offset == nethack.glyph_to_cmap(tree_glyph)
        assert tree_symdef.sym == ord("#")
        assert tree_symdef.explanation == "tree"
        assert tree_symdef.color == 2  # CLR_GREEN.
        assert str(tree_symdef) == "<nethack.symdef sym='#' explanation='tree'>"

        darkroom_glyph = nethack.GLYPH_CMAP_OFF
        assert nethack.glyph_to_cmap(darkroom_glyph) == 0
        darkroom_symdef = nethack.symdef.from_idx(0)
        assert darkroom_symdef.sym == ord(" ")
        assert darkroom_symdef.explanation == "dark part of a room"
        assert darkroom_symdef.color == 8  # NO_COLOR

    def test_glyph2tile(self):
        assert nethack.glyph2tile[nethack.GLYPH_MON_OFF] == 0
        assert nethack.glyph2tile[nethack.GLYPH_PET_OFF] == 0
        assert nethack.glyph2tile[nethack.GLYPH_DETECT_OFF] == 0

    def test_glyph_is(self):
        assert nethack.glyph_is_monster(nethack.GLYPH_MON_OFF)
        assert nethack.glyph_is_pet(nethack.GLYPH_PET_OFF)
        assert nethack.glyph_is_invisible(nethack.GLYPH_INVIS_OFF)
        assert nethack.glyph_is_detected_monster(nethack.GLYPH_DETECT_OFF)
        assert nethack.glyph_is_body(nethack.GLYPH_BODY_OFF)
        assert nethack.glyph_is_ridden_monster(nethack.GLYPH_RIDDEN_OFF)
        assert nethack.glyph_is_object(nethack.GLYPH_OBJ_OFF)
        assert nethack.glyph_is_cmap(nethack.GLYPH_CMAP_OFF)
        # No glyph_is_explode, glyph_is_zap in NH.
        assert nethack.glyph_is_swallow(nethack.GLYPH_SWALLOW_OFF)
        assert nethack.glyph_is_warning(nethack.GLYPH_WARNING_OFF)
        assert nethack.glyph_is_statue(nethack.GLYPH_STATUE_OFF)

        vec = np.array(
            [
                nethack.GLYPH_MON_OFF,
                nethack.GLYPH_PET_OFF,
                nethack.GLYPH_INVIS_OFF,
                nethack.GLYPH_DETECT_OFF,
                nethack.GLYPH_BODY_OFF,
                nethack.GLYPH_RIDDEN_OFF,
                nethack.GLYPH_OBJ_OFF,
                nethack.GLYPH_CMAP_OFF,
                nethack.GLYPH_EXPLODE_OFF,
                nethack.GLYPH_ZAP_OFF,
                nethack.GLYPH_SWALLOW_OFF,
                nethack.GLYPH_WARNING_OFF,
                nethack.GLYPH_STATUE_OFF,
            ],
            dtype=np.int32,
        )
        np.testing.assert_array_equal(
            nethack.glyph_is_monster(vec),
            np.isin(
                vec,
                [
                    nethack.GLYPH_MON_OFF,
                    nethack.GLYPH_PET_OFF,
                    nethack.GLYPH_DETECT_OFF,
                    nethack.GLYPH_RIDDEN_OFF,
                ],
            ),
        )
        np.testing.assert_array_equal(
            nethack.glyph_is_pet(vec),
            np.isin(vec, [nethack.GLYPH_PET_OFF]),
        )
        np.testing.assert_array_equal(
            nethack.glyph_is_invisible(vec),
            np.isin(vec, [nethack.GLYPH_INVIS_OFF]),
        )
        np.testing.assert_array_equal(
            nethack.glyph_is_normal_object(vec),
            np.isin(vec, [nethack.GLYPH_OBJ_OFF]),
        )
        np.testing.assert_array_equal(
            nethack.glyph_is_detected_monster(vec),
            np.isin(vec, [nethack.GLYPH_DETECT_OFF]),
        )
        np.testing.assert_array_equal(
            nethack.glyph_is_body(vec),
            np.isin(vec, [nethack.GLYPH_BODY_OFF]),
        )
        np.testing.assert_array_equal(
            nethack.glyph_is_ridden_monster(vec),
            np.isin(vec, [nethack.GLYPH_RIDDEN_OFF]),
        )
        np.testing.assert_array_equal(
            nethack.glyph_is_object(vec),
            np.isin(
                vec,
                [
                    nethack.GLYPH_BODY_OFF,
                    nethack.GLYPH_OBJ_OFF,
                    nethack.GLYPH_STATUE_OFF,
                ],
            ),
        )
        assert np.all(nethack.glyph_is_trap(vec) == 0)
        for idx in range(nethack.MAXPCHARS):  # Find an actual trap.
            if "trap" in nethack.symdef.from_idx(idx).explanation:
                assert nethack.glyph_is_trap(nethack.GLYPH_CMAP_OFF + idx)
                break
        np.testing.assert_array_equal(  # Explosions are cmaps?
            nethack.glyph_is_cmap(vec),
            np.isin(vec, [nethack.GLYPH_CMAP_OFF, nethack.GLYPH_EXPLODE_OFF]),
        )
        # No glyph_is_explode, glyph_is_zap in NH.
        np.testing.assert_array_equal(
            nethack.glyph_is_swallow(vec),
            np.isin(vec, [nethack.GLYPH_SWALLOW_OFF]),
        )
        np.testing.assert_array_equal(
            nethack.glyph_is_warning(vec),
            np.isin(vec, [nethack.GLYPH_WARNING_OFF]),
        )
        np.testing.assert_array_equal(
            nethack.glyph_is_statue(vec),
            np.isin(vec, [nethack.GLYPH_STATUE_OFF]),
        )

        # Test some non-offset value too.
        assert nethack.glyph_is_warning(
            (nethack.GLYPH_WARNING_OFF + nethack.GLYPH_STATUE_OFF) // 2
        )

    def test_glyph_to(self):
        assert np.all(
            nethack.glyph_to_mon(
                np.array(
                    [
                        nethack.GLYPH_MON_OFF,
                        nethack.GLYPH_PET_OFF,
                        nethack.GLYPH_DETECT_OFF,
                        nethack.GLYPH_RIDDEN_OFF,
                        nethack.GLYPH_STATUE_OFF,
                    ]
                )
            )
            == 0
        )

        # STATUE and CORPSE from onames.h (generated by makedefs).
        # Returned by glyph_to_obj.
        corpse = get_object("corpse").oc_name_idx
        statue = get_object("statue").oc_name_idx
        np.testing.assert_array_equal(
            nethack.glyph_to_obj(
                np.array(
                    [
                        nethack.GLYPH_BODY_OFF,
                        nethack.GLYPH_STATUE_OFF,
                        nethack.GLYPH_OBJ_OFF,
                    ]
                )
            ),
            np.array([corpse, statue, 0]),
        )

        for idx in range(nethack.MAXPCHARS):  # Find the arrow trap.
            if nethack.symdef.from_idx(idx).explanation == "arrow trap":
                np.testing.assert_array_equal(
                    nethack.glyph_to_trap(
                        np.array([nethack.GLYPH_CMAP_OFF, nethack.GLYPH_CMAP_OFF + idx])
                    ),
                    # Traps are one-indexed in defsym_to_trap as per rm.h.
                    np.array([nethack.NO_GLYPH, 1]),
                )
                break

        np.testing.assert_array_equal(
            nethack.glyph_to_cmap(
                np.array(
                    [
                        nethack.GLYPH_CMAP_OFF,
                        nethack.GLYPH_STATUE_OFF,
                    ]
                )
            ),
            np.array([0, nethack.NO_GLYPH]),
        )

        assert nethack.glyph_to_swallow(nethack.GLYPH_SWALLOW_OFF) == 0

        np.testing.assert_array_equal(
            nethack.glyph_to_warning(
                np.arange(nethack.GLYPH_WARNING_OFF, nethack.GLYPH_STATUE_OFF)
            ),
            np.arange(nethack.WARNCOUNT),
        )


class TestNethackGlanceObservation:
    @pytest.fixture
    def game(self):  # Make sure we close even on test failure.
        g = nethack.Nethack(
            playername="MonkBot-mon-hum-neu-mal",
            observation_keys=("screen_descriptions", "glyphs", "chars"),
        )
        try:
            yield g
        finally:
            g.close()

    def test_new_observation_shapes(self, game):
        screen_descriptions, glyphs, *_ = game.reset()

        assert len(screen_descriptions.shape) == 3
        rows, cols, descr_len = screen_descriptions.shape
        assert tuple(glyphs.shape) == (rows, cols)
        assert _pynethack.nethack.NLE_SCREEN_DESCRIPTION_LENGTH == descr_len

    def test_glance_descriptions(self, game):
        episodes = 6
        for _ in range(episodes):
            desc, glyphs, chars = game.reset()
            row, col = glyphs.shape
            for i in range(row):
                for j in range(col):
                    glyph = glyphs[i][j]
                    char = chars[i][j]
                    letter = chr(char)
                    glance = "".join(chr(c) for c in desc[i][j] if c != 0)
                    if char == 32:  # no text
                        assert glance == ""
                        assert (desc[i][j] == 0).all()
                    elif glyph == 2378 and letter == ".":
                        assert glance == "floor of a room"
                    elif glyph == 333 and letter == "@":  # us!
                        assert glance == "human monk called MonkBot"
                    elif glyph == 413:  # pet cat
                        assert glance == "tame kitten"
                    elif glyph == 397:  # pet dog
                        assert glance == "tame little dog"
                    elif letter in "-":  # illustrate same char, diff descrip
                        if glyph == 2378:
                            assert glance == "grave"
                        elif glyph == 2363:
                            assert glance == "wall"
                        elif glyph == 2372:
                            assert glance == "open door"


class TestNethackTerminalObservation:
    @pytest.fixture
    def game(self):  # Make sure we close even on test failure.
        g = nethack.Nethack(
            playername="MonkBot-mon-hum-neu-mal",
            observation_keys=(
                "tty_chars",
                "tty_colors",
                "tty_cursor",
                "chars",
                "colors",
            ),
        )
        try:
            yield g
        finally:
            g.close()

    def test_new_observation_shapes(self, game):
        tty_chars, tty_colors, tty_cursor, *_ = game.reset()

        assert tty_colors.shape == tty_chars.shape
        assert tty_cursor.shape == (2,)
        terminal_shape = tty_chars.shape
        assert _pynethack.nethack.NLE_TERM_LI == terminal_shape[0]
        assert _pynethack.nethack.NLE_TERM_CO == terminal_shape[1]

    def test_observations(self, game):
        tty_chars, tty_colors, *_ = game.reset()

        top_line = "".join(chr(c) for c in tty_chars[0])
        bottom_sub1_line = "".join(chr(c) for c in tty_chars[-2])
        bottom_line = "".join(chr(c) for c in tty_chars[-1])
        assert top_line.startswith(
            "Hello MonkBot, welcome to NetHack!  You are a neutral male human Monk."
        )
        assert bottom_sub1_line.startswith("MonkBot the Candidate")
        assert bottom_line.startswith("Dlvl:1")

        for c, font in zip(tty_chars.reshape(-1), tty_colors.reshape(-1)):
            if chr(c) == "@":
                assert font == 15  # BRIGHT_WHITE
            if chr(c) == " ":
                assert font == 0  # NO_COLOR

    def test_crop(self, game):
        tty_chars, tty_colors, _, chars, colors = game.reset()

        # DUNGEON is [21, 79], TTY is [24, 80]. Crop to get alignment.
        np.testing.assert_array_equal(chars, tty_chars[1:-2, :-1])
        np.testing.assert_array_equal(colors, tty_colors[1:-2, :-1])


class TestNethackMiscObservation:
    @pytest.fixture
    def game(self):  # Make sure we close even on test failure.
        g = nethack.Nethack(
            playername="MonkBot-mon-hum-neu-mal", observation_keys=("misc", "internal")
        )
        try:
            yield g
        finally:
            g.close()

    def test_misc_yn_question(self, game):
        misc, internal = game.reset()
        while misc[2]:
            (misc, internal), done = game.step(ord(" "))
            assert not done

        assert np.all(misc == 0)
        np.testing.assert_array_equal(misc, internal[1:4])

        game.step(nethack.M("p"))  # pray
        np.testing.assert_array_equal(misc, np.array([1, 0, 0]))
        np.testing.assert_array_equal(misc, internal[1:4])

        game.step(ord("n"))
        assert np.all(misc == 0)
        np.testing.assert_array_equal(misc, internal[1:4])

    def test_misc_getline(self, game):
        misc, internal = game.reset()
        while misc[2]:
            (misc, internal), done = game.step(ord(" "))
            assert not done

        assert np.all(misc == 0)
        np.testing.assert_array_equal(misc, internal[1:4])

        game.step(nethack.M("n"))  # name ..
        game.step(ord("a"))  # ... the current level
        np.testing.assert_array_equal(misc, np.array([0, 1, 0]))
        np.testing.assert_array_equal(misc, internal[1:4])

        for let in "Gehennom":
            game.step(ord(let))
            np.testing.assert_array_equal(misc, np.array([0, 1, 0]))
            np.testing.assert_array_equal(misc, internal[1:4])

        game.step(ord("\r"))
        assert np.all(misc == 0)
        np.testing.assert_array_equal(misc, internal[1:4])

    def test_misc_wait_for_space(self, game):
        misc, internal = game.reset()
        while misc[2]:
            (misc, internal), done = game.step(ord(" "))
            assert not done

        assert np.all(misc == 0)
        np.testing.assert_array_equal(misc, internal[1:4])

        game.step(ord("i"))
        np.testing.assert_array_equal(misc, np.array([0, 0, 1]))
        np.testing.assert_array_equal(misc, internal[1:4])

        game.step(ord(" "))
        assert np.all(misc == 0)
        np.testing.assert_array_equal(misc, internal[1:4])


class TestAuxillaryFunctions:
    def test_tty_render(self):
        text = ["DE", "HV"]
        chars = np.array([[ord(c) for c in line] for line in text])
        colors = np.array([[1, 2], [3, 14]])
        cursor = (0, 1)

        expected = (
            "\n\033[0;31mD\033[4m\033[0;32mE\033[0m\n\033[0;33mH\033[1;36mV\033[0m"
        )
        assert expected == nethack.tty_render(chars, colors, cursor)


class TestNethackActions:
    def test_all_ascii(self):
        for c in range(32, 127):
            if chr(c) in "%]{|}~":  # Not a NetHack command.
                continue
            nethack.action_id_to_type(c)



import gym
import minihack  # pylint: disable=unused-import
import nle  # pylint: disable=unused-import
from minihack.scripts.env_list import skip_envs_list

from nle.language_wrapper import NLELanguageWrapper


def main(nethack_env_name):
    """
    Play a NLE based environment using the nle-language-wrapper.
    """
    env = NLELanguageWrapper(
        gym.make(
            nethack_env_name,
            observation_keys=[
                "glyphs",
                "blstats",
                "tty_chars",
                "inv_letters",
                "inv_strs",
                "tty_cursor",
                "tty_colors",
            ],
        )
    )
    obsv = env.reset()
    total_reward = 0.0
    shown_help = False
    done = False

    while True:
        output = ""
        output += f"Inventory:\n{obsv['text_inventory']}\n\n"
        output += f"Stats:\n{obsv['text_blstats']}\n\n"
        output += f"Cursor:{obsv['text_cursor']}\n\n"
        output += f"Observation:\n{obsv['text_glyphs']}\n\n"
        output += f"Message:\n{obsv['text_message']}"
        if output[-1] != "\n":
            output += "\n"
        print(output)
        print("------")
        if done:
            return total_reward
        valid_action = False
        while not valid_action:
            if not shown_help:
                print(
                    'Type "instructions" anytime to see '
                    "the instructions on how to play"
                )
                print(
                    'Type "actions" anytime to see valid actions '
                    "for the current environment"
                )
                print('Type "render" anytime to render the nle environment')
                shown_help = True
            action = input("Action: ")
            if action == "actions":
                print("")
                print("Actions")
                max_key_len = max(
                    len(repr(k)) for k, _ in env.action_str_enum_map.items()
                )
                for action_str, nle_action_enum in env.action_str_enum_map.items():
                    key_str = f"{repr(action_str)}"
                    print(
                        key_str
                        + (max_key_len - len(key_str)) * " "
                        + f":{str(nle_action_enum)}"
                    )
                print("")
                continue
            if action == "instructions":
                print("")
                print("Instructions")
                print(
                    "To play, write a text action at the prompt, "
                    'e.g "north" will move you north.'
                )
                print(
                    "For the complete list of actions supported "
                    'in the current environment type "actions"'
                )
                print('To render the current display type "render"')
                print("")
                continue
            if action == ("render"):
                env.render()
                continue
            try:
                (
                    obsv,
                    reward,
                    done,
                    _,
                ) = env.step(action)
                total_reward += reward
                valid_action = True
            except ValueError as exception:
                print(exception)
    return total_reward


if __name__ == "__main__":
    while True:
        nle_env_names = [
            env_spec.id
            for env_spec in gym.envs.registry.all()
            if "MiniHack" in env_spec.id and env_spec.id not in skip_envs_list
        ]
        nle_env_names.append("NetHackChallenge-v0")
        for i, env_name in enumerate(nle_env_names):
            print(f"{i}: {env_name}")
        selected_name = int(
            input(
                f"Which base env would you like to use [0-{len(nle_env_names) - 1}]? "
            )
        )
        nle_env_name = nle_env_names[selected_name]
        total_reward = main(nle_env_name)
        print("Done!")
        print(f"Total Reward: {total_reward}")
        if input("Play again? [y,n]:") != "y":
            break



import json

import pytest  # NOQA: F401
from test_converter import getfilename
from test_db import conn  # NOQA: F401
from test_db import mockdata  # NOQA: F401

from nle import nethack

TTYRECS_TABLE_OFFSET = 0
GAMES_TABLE_OFFSET = 5
DATASETS_TABLE_OFFSET = 27 + GAMES_TABLE_OFFSET
ROOTS_TABLE_OFFSET = 2 + DATASETS_TABLE_OFFSET

TTYRECS_PATH_IDX = TTYRECS_TABLE_OFFSET + 0
TTYRECS_PART_IDX = TTYRECS_TABLE_OFFSET + 1
TTYRECS_MTIME_IDX = TTYRECS_TABLE_OFFSET + 3
TTYRECS_GAMEID_IDX = TTYRECS_TABLE_OFFSET + 4
GAMES_GAMEID_IDX = GAMES_TABLE_OFFSET + 0
GAMES_VERSION_IDX = GAMES_TABLE_OFFSET + 1
GAMES_ROLE_IDX = GAMES_TABLE_OFFSET + 12
GAMES_RACE_IDX = GAMES_TABLE_OFFSET + 13
GAMES_GEN_IDX = GAMES_TABLE_OFFSET + 14
GAMES_ALIGN_IDX = GAMES_TABLE_OFFSET + 15
GAMES_NAME_IDX = GAMES_TABLE_OFFSET + 16
GAMES_DEATH_IDX = GAMES_TABLE_OFFSET + 17
TTYREC_VERSION_IDX = ROOTS_TABLE_OFFSET + 2

DATASETS_GAMEID_IDX = DATASETS_TABLE_OFFSET + 0


class TestPopulateDB:
    def test_dump_altorg_db(self, conn):  # NOQA: F811
        # Test that populating with a mock altorg directory generates
        # the right database (effectively an integration test)
        with open(getfilename("altorg/db.json"), "r") as f:
            dump = json.load(f)

        cmd = """
        SELECT * FROM ttyrecs
        INNER JOIN games ON games.gameid=ttyrecs.gameid
        INNER JOIN datasets ON games.gameid=datasets.gameid
        INNER JOIN roots ON roots.dataset_name=datasets.dataset_name
        WHERE datasets.dataset_name='altorgtest'
        ORDER BY gameid
        """
        result = conn.execute(cmd).fetchall()
        assert len(result) == len(dump)
        # We expect repeats of the dataset generation to insert
        # games and ttyrecs in the right order into the database.
        # Since the database may already have a different dataset,
        # the gameids may be offset, so we check the relative order:
        # EG: if expected (3, 3, 4, 5, 5) actual can be (11, 11, 12, 13, 13)
        actual_gameid_offset = result[0][TTYRECS_GAMEID_IDX] - 1
        expected_gameid_offset = dump[0][TTYRECS_GAMEID_IDX] - 1
        offset = actual_gameid_offset - expected_gameid_offset
        for actual, expected in zip(result, dump):
            expected[TTYRECS_MTIME_IDX] = actual[TTYRECS_MTIME_IDX]
            assert type(actual[TTYRECS_MTIME_IDX]) == float
            expected[TTYRECS_GAMEID_IDX] += offset
            expected[GAMES_GAMEID_IDX] += offset
            expected[DATASETS_GAMEID_IDX] += offset

            assert actual == tuple(expected)

    def test_dump_nle_db(self, conn):  # NOQA: F811
        # Test that populating with an nle_data directory generates
        # the right database (effectively an integration test)

        cmd = """
        SELECT * FROM ttyrecs
        INNER JOIN games ON games.gameid=ttyrecs.gameid
        INNER JOIN datasets ON games.gameid=datasets.gameid
        INNER JOIN roots ON roots.dataset_name=datasets.dataset_name
        WHERE datasets.dataset_name='nletest'
        ORDER BY gameid
        """
        result = conn.execute(cmd).fetchall()
        endings = [
            ".0.ttyrec%i.bz2",
            ".2.ttyrec%i.bz2",
            ".4.ttyrec%i.bz2",
            ".0.ttyrec%i.bz2",
            ".2.ttyrec%i.bz2",
            ".4.ttyrec%i.bz2",
        ]
        endings = [e % nethack.TTYREC_VERSION for e in endings]
        assert len(result) == 6

        paths = []
        for i, actual in enumerate(result):
            paths.append(actual[TTYRECS_PATH_IDX])

            assert actual[TTYRECS_PATH_IDX].endswith(endings[i])
            assert actual[TTYRECS_PART_IDX] == 0
            assert actual[TTYRECS_GAMEID_IDX] == result[0][TTYRECS_GAMEID_IDX] + i
            assert actual[GAMES_VERSION_IDX] == "3.6.6"
            assert actual[GAMES_ROLE_IDX] == "Mon"
            assert actual[GAMES_RACE_IDX] == "Hum"
            assert actual[GAMES_GEN_IDX] == "Mal"
            assert actual[GAMES_ALIGN_IDX] == "Neu"
            assert actual[GAMES_NAME_IDX] == "Agent"
            assert actual[GAMES_DEATH_IDX] == "escaped"
            assert actual[TTYREC_VERSION_IDX] == nethack.TTYREC_VERSION

        assert paths == sorted(paths)



import timeit

import gym
import numpy as np
import pytest
from nle.nethack import actions as nethack_actions

from nle.language_wrapper import NLELanguageWrapper
from nle.language_wrapper.scripts import play


def test_create_env_real(real_nethack_env):
    dut = NLELanguageWrapper(real_nethack_env)
    dut.reset()


def test_env_language_action_space(real_nethack_env):
    dut = NLELanguageWrapper(real_nethack_env, use_language_action=True)
    assert isinstance(dut.action_space, gym.spaces.Space)


def test_env_discrete_action_space(real_nethack_env):
    dut = NLELanguageWrapper(real_nethack_env, use_language_action=False)
    assert isinstance(dut.action_space, gym.spaces.Discrete)


def test_env_obsv_space(real_nethack_env):
    dut = NLELanguageWrapper(real_nethack_env)
    expected_observation_space_keys = [
        "text_glyphs",
        "text_message",
        "text_blstats",
        "text_inventory",
        "text_cursor",
    ]
    assert isinstance(dut.observation_space, gym.spaces.Dict)
    for key in expected_observation_space_keys:
        assert key in dut.observation_space.spaces
    for _, value in dut.observation_space.spaces.items():
        assert isinstance(value, gym.spaces.Space)


def test_step_real(real_nethack_env):
    dut = NLELanguageWrapper(
        real_nethack_env,
    )
    dut.reset()
    dut.step("northwest")


def test_step_invalid_action(real_nethack_env):
    dut = NLELanguageWrapper(real_nethack_env)
    dut.reset()
    with pytest.raises(
        ValueError, match="is not recognized or not supported for this environment"
    ):
        dut.step("invalid action")


def test_action_actions_maps_reflect_valid_actions(fake_nle_env):
    fake_nle_env.actions = [nethack_actions.CompassDirection.N]
    dut = NLELanguageWrapper(fake_nle_env)
    assert dut.action_enum_index_map == {nethack_actions.CompassDirection.N: 0}
    assert dut.action_str_enum_map == {
        "north": nethack_actions.CompassDirection.N,
        "k": nethack_actions.CompassDirection.N,
    }


def test_step_valid_action_not_supported(real_nethack_env):
    real_nethack_env.actions = [
        action
        for action in list(real_nethack_env.actions)
        if action != nethack_actions.Command.TRAVEL
    ]

    dut = NLELanguageWrapper(real_nethack_env)
    dut.reset()
    dut.env.actions = list(dut.env.actions)
    with pytest.raises(ValueError):
        dut.step("travel")


def test_obsv_fake(fake_nle_env):
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    expected_cursor = "Yourself a monk"
    expected_inventory = (
        "a: a blessed +1 quarterstaff (weapon in hands)\n"
        "b: an uncursed +0 cloak of magic resistance (being worn)"
    )
    expected_message = (
        "Hello Agent, welcome to NetHack!  You are a neutral gnomish Caveman."
    )
    expected_obsv = (
        "tame little dog far westsouthwest\n"
        "fountain near northnorthwest\n"
        "dark area very near north\n"
        "vertical closed door adjacent north\n"
        "dark area adjacent northeast, east, southeast, "
        "south, southwest, west, and northwest"
    )
    expected_text_blstats = (
        "Strength: 1/19\n"
        "Dexterity: 4\n"
        "Constitution: 5\n"
        "Intelligence: 6\n"
        "Wisdom: 7\n"
        "Charisma: 8\n"
        "Depth: 9\n"
        "Gold: 10\n"
        "HP: 11/12\n"
        "Energy: 1/13\n"
        "AC: 1\n"
        "XP: 7/10\n"
        "Time: 5\n"
        "Position: 66|14\n"
        "Hunger: Fainting\n"
        "Monster Level: 23\n"
        "Encumbrance: Stressed\n"
        "Dungeon Number: 21\n"
        "Level Number: 10\n"
        "Score: 123\n"
        "Alignment: Chaotic\n"
        "Condition: Food Poisoning Slimed Stoned"
    )

    assert obsv["text_glyphs"] == expected_obsv
    assert obsv["text_message"] == expected_message
    assert obsv["text_inventory"] == expected_inventory
    assert obsv["text_blstats"] == expected_text_blstats
    assert obsv["text_cursor"] == expected_cursor


def test_blstats_condition_none(fake_nle_env):
    # Set condition to None.
    fake_nle_env.reset.return_value["blstats"][25] = 0

    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()

    expected_condition = "Condition: None"

    assert expected_condition in obsv["text_blstats"]


def test_blstats_condition_flying(fake_nle_env):
    # Set condition to Flying.
    fake_nle_env.reset.return_value["blstats"][25] = 2048

    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()

    expected_condition = "Condition: Flying"

    assert expected_condition in obsv["text_blstats"]


def test_multiple_obsv_fake(fake_nethack_multiple_monsters_env):
    dut = NLELanguageWrapper(fake_nethack_multiple_monsters_env)
    obsv = dut.reset()
    expected_inventory = (
        "a: a blessed +1 quarterstaff (weapon in hands)\n"
        "b: an uncursed +0 cloak of magic resistance (being worn)"
    )
    expected_message = (
        "Hello Agent, welcome to NetHack!  You are a neutral gnomish Caveman."
    )
    expected_obsv = (
        "tame little dog very far west\n"
        "tame little dogs far westsouthwest\n"
        "fountains near northnorthwest\n"
        "dark area very near north\n"
        "vertical closed door adjacent north\n"
        "dark area adjacent northeast, east, "
        "southeast, south, southwest, west, and northwest"
    )
    assert obsv["text_glyphs"] == expected_obsv
    assert obsv["text_message"] == expected_message
    assert obsv["text_inventory"] == expected_inventory


def test_step_fake(fake_nle_env):
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    expected_inventory = (
        "a: a blessed +1 quarterstaff (weapon in hands)\n"
        "b: an uncursed +0 cloak of magic resistance (being worn)"
    )
    expected_message = (
        "Hello Agent, welcome to NetHack!  You are a neutral gnomish Caveman."
    )
    expected_obsv = (
        "tame little dog far westsouthwest\n"
        "fountain near northnorthwest\n"
        "dark area very near north\n"
        "vertical closed door adjacent north\n"
        "dark area adjacent northeast, east, "
        "southeast, south, southwest, west, and northwest"
    )
    obsv, reward, done, info = dut.step("north")
    assert obsv["text_glyphs"] == expected_obsv
    assert obsv["text_message"] == expected_message
    assert obsv["text_inventory"] == expected_inventory
    assert reward == 0.0
    assert not done
    assert info is None


def test_statue(fake_nle_env):
    glyphs = np.full((21, 79), 2378)  # Room floor
    glyphs[14, 67] = 5653  # kobold statue
    fake_nle_env.reset.return_value["glyphs"] = glyphs
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    assert obsv["text_glyphs"] == "kobold statue adjacent east"


def test_warning(fake_nle_env):
    glyphs = np.full((21, 79), 2378)  # Room floor
    glyphs[14, 67] = 5593  # unknown creature causing you alarm adjacent east
    fake_nle_env.reset.return_value["glyphs"] = glyphs
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    assert obsv["text_glyphs"] == "unknown creature causing you alarm adjacent east"


def test_swallow(fake_nle_env):
    glyphs = np.full((21, 79), 2378)  # Room floor
    glyphs[14, 67] = 5332  # swallow
    fake_nle_env.reset.return_value["glyphs"] = glyphs
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    assert obsv["text_glyphs"] == "swallow bottom right adjacent east"


def test_zap_beam(fake_nle_env):
    glyphs = np.full((21, 79), 2378)  # Room floor
    glyphs[14, 67] = 2528  # right slant beam
    fake_nle_env.reset.return_value["glyphs"] = glyphs
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    assert obsv["text_glyphs"] == "right slant zap beam adjacent east"


def test_explosion(fake_nle_env):
    glyphs = np.full((21, 79), 2378)  # Room floor
    glyphs[14, 67] = 2460  # explosion middle right
    fake_nle_env.reset.return_value["glyphs"] = glyphs
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    assert obsv["text_glyphs"] == "explosion middle right adjacent east"


def test_illegal_object(fake_nle_env):
    glyphs = np.full((21, 79), 2378)  # Room floor
    glyphs[14, 67] = 1906  # strange object
    fake_nle_env.reset.return_value["glyphs"] = glyphs
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    assert obsv["text_glyphs"] == "strange object adjacent east"


def test_weapon(fake_nle_env):
    glyphs = np.full((21, 79), 2378)  # Room floor
    glyphs[14, 67] = 1917  # elven spear
    fake_nle_env.reset.return_value["glyphs"] = glyphs
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    assert obsv["text_glyphs"] == "elven spear adjacent east"


def test_armour(fake_nle_env):
    glyphs = np.full((21, 79), 2378)  # Room floor
    glyphs[14, 67] = 1988  # gray dragon scale mail
    fake_nle_env.reset.return_value["glyphs"] = glyphs
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    assert obsv["text_glyphs"] == "gray dragon scale mail adjacent east"


def test_ring(fake_nle_env):
    glyphs = np.full((21, 79), 2378)  # Room floor
    glyphs[14, 67] = 2068  # sapphire ring
    fake_nle_env.reset.return_value["glyphs"] = glyphs
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    assert obsv["text_glyphs"] == "sapphire ring adjacent east"


def test_amulet(fake_nle_env):
    glyphs = np.full((21, 79), 2378)  # Room floor
    glyphs[14, 67] = 2091  # hexagonal amulet
    fake_nle_env.reset.return_value["glyphs"] = glyphs
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    assert obsv["text_glyphs"] == "hexagonal amulet adjacent east"


def test_tool(fake_nle_env):
    glyphs = np.full((21, 79), 2378)  # Room floor
    glyphs[14, 67] = 2123  # magic marker
    fake_nle_env.reset.return_value["glyphs"] = glyphs
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    assert obsv["text_glyphs"] == "magic marker adjacent east"


def test_food(fake_nle_env):
    glyphs = np.full((21, 79), 2378)  # Room floor
    glyphs[14, 67] = 2163  # carrot
    fake_nle_env.reset.return_value["glyphs"] = glyphs
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    assert obsv["text_glyphs"] == "carrot adjacent east"


def test_potion(fake_nle_env):
    glyphs = np.full((21, 79), 2378)  # Room floor
    glyphs[14, 67] = 2187  # magenta potion
    fake_nle_env.reset.return_value["glyphs"] = glyphs
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    assert obsv["text_glyphs"] == "magenta potion adjacent east"


def test_scroll(fake_nle_env):
    glyphs = np.full((21, 79), 2378)  # Room floor
    glyphs[14, 67] = 2206  # scroll labeled NR 9
    fake_nle_env.reset.return_value["glyphs"] = glyphs
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    assert obsv["text_glyphs"] == "scroll labeled NR 9 adjacent east"


def test_spellbook(fake_nle_env):
    glyphs = np.full((21, 79), 2378)  # Room floor
    glyphs[14, 67] = 2250  # mottled spellbook
    fake_nle_env.reset.return_value["glyphs"] = glyphs
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    assert obsv["text_glyphs"] == "mottled spellbook adjacent east"


def test_wand(fake_nle_env):
    glyphs = np.full((21, 79), 2378)  # Room floor
    glyphs[14, 67] = 2291  # crystal wand
    fake_nle_env.reset.return_value["glyphs"] = glyphs
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    assert obsv["text_glyphs"] == "crystal wand adjacent east"


def test_coin(fake_nle_env):
    glyphs = np.full((21, 79), 2378)  # Room floor
    glyphs[14, 67] = 2316  # gold piece
    fake_nle_env.reset.return_value["glyphs"] = glyphs
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    assert obsv["text_glyphs"] == "gold piece adjacent east"


def test_gem(fake_nle_env):
    glyphs = np.full((21, 79), 2378)  # Room floor
    glyphs[14, 67] = 2333  # violet amethyst
    fake_nle_env.reset.return_value["glyphs"] = glyphs
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    assert obsv["text_glyphs"] == "violet amethyst adjacent east"


def test_rock(fake_nle_env):
    glyphs = np.full((21, 79), 2378)  # Room floor
    glyphs[14, 67] = 2353  # boulder
    fake_nle_env.reset.return_value["glyphs"] = glyphs
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    assert obsv["text_glyphs"] == "boulder adjacent east"


def test_ball(fake_nle_env):
    glyphs = np.full((21, 79), 2378)  # Room floor
    glyphs[14, 67] = 2355  # heavy iron ball
    fake_nle_env.reset.return_value["glyphs"] = glyphs
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    assert obsv["text_glyphs"] == "heavy iron ball adjacent east"


def test_chain(fake_nle_env):
    glyphs = np.full((21, 79), 2378)  # Room floor
    glyphs[14, 67] = 2356  # iron chain
    fake_nle_env.reset.return_value["glyphs"] = glyphs
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    assert obsv["text_glyphs"] == "iron chain adjacent east"


def test_venom(fake_nle_env):
    glyphs = np.full((21, 79), 2378)  # Room floor
    glyphs[14, 67] = 2358  # splash of acid venom
    fake_nle_env.reset.return_value["glyphs"] = glyphs
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    assert obsv["text_glyphs"] == "splash of acid venom adjacent east"


def test_ridden(fake_nle_env):
    glyphs = np.full((21, 79), 2378)  # Room floor
    glyphs[14, 67] = 1611  # ridden mastodon
    fake_nle_env.reset.return_value["glyphs"] = glyphs
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    assert obsv["text_glyphs"] == "ridden mastodon adjacent east"


def test_corpse(fake_nle_env):
    glyphs = np.full((21, 79), 2378)  # Room floor
    glyphs[14, 67] = 1226  # leocrotta corpse
    fake_nle_env.reset.return_value["glyphs"] = glyphs
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    assert obsv["text_glyphs"] == "leocrotta corpse adjacent east"


def test_invisible(fake_nle_env):
    glyphs = np.full((21, 79), 2378)  # Room floor
    glyphs[14, 67] = 762  # invisible creature
    fake_nle_env.reset.return_value["glyphs"] = glyphs
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    assert obsv["text_glyphs"] == "invisible creature adjacent east"


def test_detected(fake_nle_env):
    glyphs = np.full((21, 79), 2378)  # Room floor
    glyphs[14, 67] = 917  # detected water elemental
    fake_nle_env.reset.return_value["glyphs"] = glyphs
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    assert obsv["text_glyphs"] == "detected water elemental adjacent east"


def test_tame(fake_nle_env):
    glyphs = np.full((21, 79), 2378)  # Room floor
    glyphs[14, 67] = 613  # tame yeti
    fake_nle_env.reset.return_value["glyphs"] = glyphs
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    assert obsv["text_glyphs"] == "tame yeti adjacent east"


def test_monster(fake_nle_env):
    glyphs = np.full((21, 79), 2378)  # Room floor
    glyphs[14, 67] = 255  # iron golem
    fake_nle_env.reset.return_value["glyphs"] = glyphs
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    assert obsv["text_glyphs"] == "iron golem adjacent east"


def test_plural_end_ey(fake_nle_env):
    glyphs = np.full((21, 79), 2378)  # Room floor
    glyphs[14, 72] = 229  # monkey 1'
    glyphs[14, 73] = 229  # monkey 2
    fake_nle_env.reset.return_value["glyphs"] = glyphs
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    assert obsv["text_glyphs"] == "monkeys far east"


def test_plural_end_y(fake_nle_env):
    glyphs = np.full((21, 79), 2378)  # Room floor
    glyphs[14, 72] = 99  # pony 1'
    glyphs[14, 73] = 99  # pony 2
    fake_nle_env.reset.return_value["glyphs"] = glyphs
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    assert obsv["text_glyphs"] == "ponies far east"


def test_plural_default(fake_nle_env):
    glyphs = np.full((21, 79), 2378)  # Room floor
    glyphs[14, 72] = 341  # wizard 1'
    glyphs[14, 73] = 341  # wizard 2
    fake_nle_env.reset.return_value["glyphs"] = glyphs
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    assert obsv["text_glyphs"] == "wizards far east"


def test_plural_end_s(fake_nle_env):
    glyphs = np.full((21, 79), 2378)  # Room floor
    glyphs[14, 72] = 335  # priestess 1'
    glyphs[14, 73] = 335  # priestess 2
    fake_nle_env.reset.return_value["glyphs"] = glyphs
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    assert obsv["text_glyphs"] == "priestesses far east"


def test_plural_end_f(fake_nle_env):
    glyphs = np.full((21, 79), 2378)  # Room floor
    glyphs[14, 72] = 260  # elf 1'
    glyphs[14, 73] = 260  # elf 2
    fake_nle_env.reset.return_value["glyphs"] = glyphs
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    assert obsv["text_glyphs"] == "elves far east"


def test_plural_end_ff(fake_nle_env):
    glyphs = np.full((21, 79), 2378)  # Room floor
    glyphs[14, 72] = 1967  # quarterstaff 1'
    glyphs[14, 73] = 1967  # quarterstaff 2
    fake_nle_env.reset.return_value["glyphs"] = glyphs
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    assert obsv["text_glyphs"] == "quarterstaves far east"


def test_plural_lava(fake_nle_env):
    glyphs = np.full((21, 79), 2378)  # Room floor
    glyphs[14, 72] = 2393  # lava 1
    glyphs[14, 73] = 2393  # lava 2
    fake_nle_env.reset.return_value["glyphs"] = glyphs
    dut = NLELanguageWrapper(fake_nle_env)
    obsv = dut.reset()
    assert obsv["text_glyphs"] == "area of lava far east"


def test_wrapper_only_works_with_nle_envs():
    with pytest.raises(AssertionError, match=r"Only NLE environments are supported"):
        NLELanguageWrapper(gym.make("CartPole-v1"))


def test_wrapper_requires_all_keys(real_nethack_env):
    real_nethack_env.observation_space.spaces.pop("glyphs")
    with pytest.raises(
        AssertionError,
        match=r"NLE environment missing required obsv key\(s\): {'glyphs'}",
    ):
        NLELanguageWrapper(real_nethack_env)


def test_play(mocker):
    with mocker.patch("builtins.input", return_value="north"):
        play.main("NetHackChallenge-v0")


def test_time_reset(real_nethack_env):
    dut = NLELanguageWrapper(real_nethack_env)
    baseline_runtime = min(timeit.repeat(real_nethack_env.reset, number=100, repeat=10))
    runtime = min(timeit.repeat(dut.reset, number=100, repeat=10))
    relative_slowdown = runtime / baseline_runtime
    # Resetting takes a lot more time than stepping so the overhead is a lot less.
    assert relative_slowdown < 1.2


def test_time_step(real_nethack_env):
    wait_action = 18
    dut = NLELanguageWrapper(real_nethack_env)
    dut.reset()
    baseline_runtime = min(
        timeit.repeat(lambda: real_nethack_env.step(wait_action), number=100, repeat=10)
    )
    runtime = min(timeit.repeat(lambda: dut.step("wait"), number=100, repeat=10))
    relative_slowdown = runtime / baseline_runtime
    assert relative_slowdown < 3.0



import numpy as np


def str_to_1d(string):
    """Convert a string to equivalent numpy array
    Args:
        string(str): string to be converted
    Returns:
        (numpy.ndarray): 1-D uint8 values for unicode string input.
    """
    output = np.array(bytearray(string.encode("utf-8")))
    return output


def strs_to_2d(strings, fill_value=0):
    """Convert a list of strings to equivalent 2d numpy array
    Args:
        strings(List[str]): string to be converted
    Returns:
        (numpy.ndarray): 2-D uint8 values for unicode strings input.
    """
    arrays = []
    max_len = 0
    for string in strings:
        arrays.append(str_to_1d(string))
        if len(string) > max_len:
            max_len = len(string)
    output = np.full((len(arrays), max_len), fill_value=fill_value, dtype=np.uint8)
    for i, array in enumerate(arrays):
        output[i, : len(array)] = array
    return output



import torch
from sample_factory.algorithms.appo.model_utils import EncoderBase
from sample_factory.algorithms.appo.model_utils import register_custom_encoder
from transformers import RobertaConfig
from transformers import RobertaModel


class NLELanguageTransformerEncoder(EncoderBase):
    def __init__(self, cfg, obs_space, timing):
        super().__init__(cfg, timing)
        config = RobertaConfig(
            attention_probs_dropout_prob=0.0,
            bos_token_id=0,
            classifier_dropout=None,
            eos_token_id=2,
            hidden_act="gelu",
            hidden_dropout_prob=0.0,
            hidden_size=cfg.transformer_hidden_size,
            initializer_range=0.02,
            intermediate_size=cfg.transformer_hidden_size,
            layer_norm_eps=1e-05,
            max_position_embeddings=obs_space.spaces["input_ids"].shape[0]
            + 2,  # Roberta requires max sequence length + 2.
            model_type="roberta",
            num_attention_heads=cfg.transformer_attention_heads,
            num_hidden_layers=cfg.transformer_hidden_layers,
            pad_token_id=1,
            position_embedding_type="absolute",
            transformers_version="4.17.0",
            type_vocab_size=1,
            use_cache=False,
            vocab_size=50265,
        )
        self.model = RobertaModel(config=config)
        self.encoder_out_size = self.model.config.hidden_size

    def device_and_type_for_input_tensor(
        self, input_tensor_name
    ):  # pylint: disable=['unused-argument']
        return "cuda", torch.int32  # pylint: disable=['no-member']

    def forward(self, obs_dict):
        input_ids = obs_dict["input_ids"]
        attention_mask = obs_dict["attention_mask"]
        # Input transformation to allow for sample factory enjoy
        if len(input_ids.shape) == 3:
            input_ids = input_ids.squeeze(0)
            attention_mask = attention_mask.squeeze(0)
        if input_ids.dtype == torch.float32:  # pylint: disable=['no-member']
            input_ids = input_ids.long()
            attention_mask = attention_mask.long()
        output = self.model(input_ids=input_ids, attention_mask=attention_mask)
        return output.last_hidden_state[:, 0]


register_custom_encoder(
    "nle_language_transformer_encoder", NLELanguageTransformerEncoder
)



import bz2
import concurrent.futures as futures
import contextlib

import numpy as np
import pytest
from test_converter import COLSROWS
from test_converter import FINALFRAME
from test_converter import FINALFRAMECOLORS
from test_converter import TIMESTAMPS
from test_converter import getfilename
from test_db import conn  # noqa: F401
from test_db import mockdata  # noqa: F401

from nle.dataset import dataset
from nle.dataset import db


class TestDataset:
    @pytest.fixture
    def db_exists(self, conn):  # noqa: F811
        """Loading this module ensure the ttyrec.db can be found by TtyrecDataset"""
        del conn

    @pytest.fixture(params=["nopool", "threadpool"])
    def pool(self, request):
        if request.param == "threadpool":
            cm = futures.ThreadPoolExecutor
        else:
            try:
                cm = contextlib.nullcontext
            except AttributeError:
                # `nullcontext`` is Python 3.7+ feature.
                @contextlib.contextmanager
                def nullcontext(*args, **kwargs):
                    yield

                cm = lambda: nullcontext()
        with cm() as tp:
            yield tp

    def test_setup(self, conn):  # noqa: F811
        files = [db.get_row(f"{i+1}", conn=conn)[1] for i in range(9)]
        names = ["aaa", "bbb", "ccc"]
        assert files == [f"{a}/{b}.ttyrec.bz2" for a in names for b in names]

    def test_dataset_gameids(self, db_exists, pool):
        gameids = np.random.choice(7, 7, replace=False) + 1
        data = dataset.TtyrecDataset(
            "basictest", batch_size=7, threadpool=pool, gameids=gameids
        )
        mb = next(iter(data))
        files = mb["gameids"]
        assert len(mb) == 6

        np.testing.assert_array_equal(np.unique(files[:, 0]), np.arange(1, 8))

    def test_minibatches(self, db_exists, pool):
        data = dataset.TtyrecDataset(
            "basictest",
            seq_length=50,
            batch_size=4,
            threadpool=pool,
            gameids=range(1, 8),
            shuffle=False,
        )
        # starting gameids = [TTYREC, TTYREC, TTYREC, TTYREC2]

        mb = next(iter(data))
        for name, array in mb.items():
            if name in ("gameids",):
                continue
            # Test first three rows are the same, and differ from from fourth
            np.testing.assert_array_equal(array[0], array[1])
            np.testing.assert_array_equal(array[0], array[2])
            np.testing.assert_raises(
                AssertionError, np.testing.assert_array_equal, array[0], array[3]
            )

        # Check reseting occured
        reset = np.where(mb["done"][3] == 1)[0][0]
        assert reset == 31

        # Check the data at location is the same. Note reset occurs for batch 4
        seq = 10
        for name, array in mb.items():
            if name in ("done", "gameids"):
                continue
            np.testing.assert_array_equal(array[3][:seq], array[3][reset : reset + seq])

        # No leading 1s
        assert (mb["done"][:, 0] == 0).all()

    def test_get_ttyrec(self, db_exists, pool):
        data = dataset.TtyrecDataset(
            "basictest",
            seq_length=100,
            batch_size=1,
            gameids=[4, 5],
            threadpool=pool,
            shuffle=False,
        )

        mb = next(iter(data))
        reset = np.where(mb["done"][0] == 1)[0][0]
        assert reset == 31

        for chunk_size in [5, 50]:
            chunks = data.get_ttyrec(4, chunk_size=chunk_size)

            groups = {k: [] for k in mb.keys()}
            for c in chunks:
                for k in c:
                    groups[k].append(c[k])
            concat_chunks = {k: np.concatenate(t, axis=1) for k, t in groups.items()}
            for k in mb.keys():
                c = concat_chunks[k]
                m = mb[k][0]
                np.testing.assert_array_equal(c[0, :reset], m[:reset])
                np.testing.assert_equal(c[0, reset:], 0)

    def test_char_frame(self, db_exists, pool):
        # gameids [1,2,3] are all the same tty_rec, rowid 4 is different
        gameids = [1, 2, 3, 4, 1, 2, 3, 4]
        seq_length = 20
        data = dataset.TtyrecDataset(
            "basictest",
            seq_length=seq_length,
            rows=25,
            batch_size=3,
            gameids=gameids,
            threadpool=pool,
            shuffle=False,
        )

        with open(getfilename(FINALFRAME), "r") as f:
            lines = [line.rstrip() for line in f.readlines()]

        with open(getfilename(FINALFRAMECOLORS), "r") as f:
            colorlines = [line.rstrip().split(",") for line in f.readlines()]

        with open(getfilename(COLSROWS)) as f:
            colsrows = [tuple(int(i) for i in line.split()) for line in f]

        with bz2.BZ2File(getfilename(TIMESTAMPS)) as f:
            saved_timestamps = [float(line) * 1e6 for line in f]

        i = 0
        for mb in data:
            chars, colors, curs, ts, done, _ = mb.values()
            batch_ids, seq_ids = np.where(done == 1)
            seq = seq_length if not seq_ids.any() else seq_ids[0]
            for j in range(seq):
                np.testing.assert_array_equal(colsrows[i][0], curs[:, j, 1])
                np.testing.assert_array_equal(colsrows[i][1], curs[:, j, 0])
                np.testing.assert_array_almost_equal(
                    saved_timestamps[i], ts[:, j], decimal=0
                )

                i += 1
            if seq_ids.any():
                np.testing.assert_array_equal(batch_ids, np.arange(3))
                np.testing.assert_array_equal(seq_ids, seq_ids[0])

                final_frame = chars[0][seq_ids[0] - 1]
                char_frame = [
                    row.tobytes().decode("utf-8").rstrip() for row in final_frame
                ]
                color_frame = colors[0][seq_ids[0] - 1]

                for actual, expected in zip(char_frame, lines):
                    assert actual == expected
                for actual, expected in zip(color_frame, colorlines):
                    for col_a, col_e in zip(actual, expected):
                        assert col_a == int(col_e)
                break
            if i > 2431:
                raise AssertionError  # this should have ended!

    def test_start_and_end(self, db_exists):
        # gameids [1,2,3] are all the same tty_rec, rowid 4 is different
        data_1 = dataset.TtyrecDataset(
            "basictest",
            seq_length=100,
            batch_size=3,
            gameids=range(1, 4),
            shuffle=False,
        )
        data_2 = dataset.TtyrecDataset(
            "basictest", seq_length=100, batch_size=1, gameids=[4, 4, 4], shuffle=False
        )

        mb1 = next(iter(data_1))
        mb2 = next(iter(data_2))

        for k in mb1:
            if k == "gameids":
                continue
            array1 = mb1[k]
            array2 = mb2[k]
            # Test first three rows are the same, and differ from from fourth
            np.testing.assert_array_equal(array1[0], array1[1])
            np.testing.assert_array_equal(array1[0], array1[2])
            np.testing.assert_raises(
                AssertionError, np.testing.assert_array_equal, array1[0], array2[0]
            )

    @pytest.mark.parametrize("batch_size", [2, 5, 7])
    def test_no_looping(self, db_exists, pool, batch_size):
        # We expect as many done as (len(gameids) - batchsize)
        data = dataset.TtyrecDataset(
            "basictest",
            seq_length=100,
            batch_size=batch_size,
            threadpool=pool,
            gameids=range(1, 8),
        )
        done = 0
        for mb in data:
            done += np.sum(mb["done"] == 1)
            if done > 7 - batch_size:
                break
        assert done == 7 - batch_size

    @pytest.mark.parametrize("batch_size", [1, 3, 6])
    def test_shuffle(self, db_exists, batch_size):
        data = dataset.TtyrecDataset(
            "basictest",
            seq_length=1000,
            batch_size=batch_size,
            threadpool=None,
            gameids=(1, 2, 3, 4, 5, 6, 7, 6, 5, 4, 3, 2, 1),
            shuffle=False,
        )

        def get_data():
            """Get all the data from minibatches and concat into full data batches"""
            mbs = []
            gameids = []
            for mb in data:
                mbs.append([t.copy() for t in mb.values()])
                gameids.append(mb["gameids"].copy())
            return [
                np.concatenate(array_list, axis=1) for array_list in zip(*mbs)
            ], np.concatenate(gameids)

        b1, g1 = get_data()
        b2, g2 = get_data()
        for a1, a2 in zip(b1, b2):
            np.testing.assert_array_equal(a1, a2)
        np.testing.assert_array_equal(g1, g2)

        data.shuffle = True
        b3, g3 = get_data()
        b4, g4 = get_data()
        # Note: the last tensor in the batch is the gameid
        np.testing.assert_raises(AssertionError, np.testing.assert_array_equal, g3, g2)
        np.testing.assert_raises(AssertionError, np.testing.assert_array_equal, g3, g4)

        data.shuffle = False
        b5, g5 = get_data()
        for a1, a5 in zip(b1, b5):
            np.testing.assert_array_equal(a1, a5)
        np.testing.assert_array_equal(g1, g5)

    def test_sql(self, db_exists, pool):

        sql = """
            SELECT gameid FROM datasets
            WHERE gameid >=6 AND dataset_name='basictest'
            """
        data = dataset.TtyrecDataset(
            "basictest",
            seq_length=100,
            batch_size=2,
            threadpool=pool,
            shuffle=False,
            subselect_sql=sql,
        )

        data2 = dataset.TtyrecDataset(
            "basictest",
            seq_length=100,
            batch_size=2,
            threadpool=pool,
            gameids=[6, 7],
            shuffle=False,
        )
        data.populate_metadata()
        data2.populate_metadata()
        assert len(data.get_meta(6)) == 27
        assert data.get_meta(6) == data2.get_meta(6)

        for _, (mb1, mb2) in enumerate(zip(data, data2)):
            for k in mb1.keys():
                np.testing.assert_array_equal(mb1[k], mb2[k])

    def test_multipart_game(self, db_exists, pool):
        # This test selects a multipart game

        # Use sql select all of user "ccc" ttyrecs (3 ttyrecs, gameid 7) as one game
        # eg: [(7, 0, /path/to/A), (7, 1, /path/to/B), (7, 2, /path/to/C)]
        sql1 = "SELECT gameid FROM games WHERE games.name='ccc'"
        data1 = dataset.TtyrecDataset(
            "basictest",
            seq_length=100,
            batch_size=1,
            threadpool=pool,
            subselect_sql=sql1,
        )

        # Hack data2 to treat of user "ccc" ttyrecs (7, 8, 9) as separate games
        # eg: [(0, 7, /path/to/A), (1, 7, /path/to/B), (2, 7, /path/to/C)]
        data2 = dataset.TtyrecDataset(
            "basictest",
            seq_length=100,
            batch_size=1,
            threadpool=pool,
            subselect_sql=sql1,
        )
        ccc_gameid = data2._gameids[0]
        data2._games[1000] = [data2._games[ccc_gameid][0]]
        data2._games[2000] = [data2._games[ccc_gameid][1]]
        data2._games[3000] = [data2._games[ccc_gameid][2]]
        data2._gameids = [1000, 2000, 3000]

        for _, (mb1, mb2) in enumerate(zip(data1, data2)):
            for k in mb1.keys():
                if k == "gameids":
                    assert np.all(np.isin(mb1[k], [0, 7]))
                elif k == "done":
                    np.testing.assert_equal(mb1[k], np.zeros_like(mb1[k]))
                else:
                    np.testing.assert_array_equal(mb1[k], mb2[k])

    def test_dataset_metadata(self, db_exists, pool):
        # Test we can retrieve metadata from database and access it later
        # Everything after the gameid, path should be stored as a list in metadata
        sql1 = """
            SELECT games.gameid FROM games
            INNER JOIN datasets
            ON datasets.gameid = games.gameid
            WHERE datasets.dataset_name='basictest'
            AND games.name="ccc"
            """
        data1 = dataset.TtyrecDataset(
            "basictest",
            seq_length=100,
            batch_size=1,
            threadpool=pool,
            subselect_sql=sql1,
        )
        data1.populate_metadata()
        gameids = next(iter(data1))["gameids"]
        for _ in [7, 8, 9]:
            rowid = gameids[0][0]
            # These are SQLiteRows, which can be accessed
            # lists or dictionaries.
            assert data1.get_meta(rowid)[17] == "ascended"
            assert data1.get_meta(rowid)["death"] == "ascended"
            assert data1.get_meta(rowid)[2] == 999
            assert data1.get_meta(rowid)["points"] == 999



from functools import lru_cache

import gym
import numpy as np
import torch
from sample_factory.envs.env_registry import global_env_registry
from transformers import RobertaTokenizerFast

from nle.language_wrapper import NLELanguageWrapper


class SampleFactoryNLELanguageEnv(gym.Env):
    LRU_CACHE_SIZE = 1000

    def __init__(self, cfg):
        self.cfg = cfg
        self.observation_space = gym.spaces.Dict()
        self.observation_space.spaces["obs"] = gym.spaces.Box(
            0, 1000000, shape=(1,), dtype=np.int32
        )
        self.observation_space.spaces["input_ids"] = gym.spaces.Box(
            0, 1000000, shape=(self.cfg["max_token_length"],), dtype=np.int32
        )
        self.observation_space.spaces["attention_mask"] = gym.spaces.Box(
            0, 1, shape=(self.cfg["max_token_length"],), dtype=np.int32
        )
        self.nle_env = gym.make(self.cfg["nle_env_name"])
        self.env = NLELanguageWrapper(self.nle_env, use_language_action=False)
        self.action_space = self.env.action_space
        self.tokenizer = RobertaTokenizerFast.from_pretrained(
            "distilroberta-base", truncation_side="left"
        )

    # We use caching to avoid re-tokenizing observations that are already seen.
    @lru_cache(maxsize=LRU_CACHE_SIZE)
    def _tokenize(self, str_obsv):
        tokens = self.tokenizer(
            str_obsv,
            return_tensors="pt",
            padding="max_length",
            truncation=True,
            max_length=self.cfg["max_token_length"],
        )
        # Sample factory insists on normalizing obs key.
        tokens.data["obs"] = torch.zeros(1)
        return tokens.data

    def _convert_obsv_to_str(self, obsv):
        text_obsv = ""
        text_obsv += f"Inventory:\n{obsv['text_inventory']}\n\n"
        text_obsv += f"Stats:\n{obsv['text_blstats']}\n\n"
        text_obsv += f"Cursor:\n{obsv['text_cursor']}\n\n"
        text_obsv += f"Stats:\n{obsv['text_glyphs']}\n\n"
        text_obsv += f"Message:\n{obsv['text_message']}"
        return text_obsv

    def reset(self, **kwargs):
        return self._tokenize(self._convert_obsv_to_str(self.env.reset(**kwargs)))

    def step(self, action):
        obsv, reward, done, info = self.env.step(action)
        tokenized_obsv = self._tokenize(self._convert_obsv_to_str(obsv))
        return tokenized_obsv, reward, done, info

    def seed(self, *args):  # pylint: disable=['unused-argument']
        # Nethack does not allow seeding
        return

    def render(self, *args, **kwargs):  # pylint: disable=['unused-argument']
        self.env.render()


def make_custom_env_func(
    full_env_name, cfg=None, env_config=None
):  # pylint: disable=['unused-argument']
    env = SampleFactoryNLELanguageEnv(cfg)
    return env


global_env_registry().register_env(
    env_name_prefix="nle_language_env",
    make_env_func=make_custom_env_func,
)
